#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################


"""Example on using a tpg.TaurusTrendSet and some related tools
on a pure pyqtgraph plot"""

if __name__ == "__main__":
    import sys
    from taurus.qt.qtgui.application import TaurusApplication
    from taurus_pyqtgraph import (
        TaurusTrendSet,
        DateAxisItem,
        XAutoPanTool,
        ForcedReadTool,
    )
    import pyqtgraph as pg

    from taurus.core.taurusmanager import TaurusManager

    taurusM = TaurusManager()
    taurusM.changeDefaultPollingPeriod(1000)  # ms

    app = TaurusApplication()

    # Add a date-time X axis
    axis = DateAxisItem(orientation="bottom")
    w = pg.PlotWidget()
    axis.attachToPlotItem(w.getPlotItem())

    # Add the auto-pan ("oscilloscope mode") tool
    autopan = XAutoPanTool()
    autopan.attachToPlotItem(w.getPlotItem())

    # Add Forced-read tool
    fr = ForcedReadTool(w, period=1000)
    fr.attachToPlotItem(w.getPlotItem())

    # add legend the legend tool
    w.addLegend()

    # adding a taurus data item...
    c2 = TaurusTrendSet(name="foo")
    c2.setModel("eval:rand(2)")
    w.addItem(c2)

    w.show()

    sys.exit(app.exec_())



import csv
from datetime import datetime
import itertools
import os

import numpy as np
from pyqtgraph.exporters.CSVExporter import Exporter, PlotItem
from pyqtgraph.Qt import QtWidgets
from pyqtgraph.widgets.FileDialog import FileDialog


LastExportDirectory = None


class Taurus4TextExporter(Exporter):
    """
    Attempt to imitate the data export format used for taurustrend in Taurus 4
    (and maybe older). Tab separated values, with (somewhat) ISO timestamps
    and a two line info header

    # DATASET= "tango://tangohost:10000/sys/tg_test/1/double_scalar[0]"
    # SNAPSHOT_TIME= 2024-01-31_10:22:59.803364
    2024-01-31_10:21:50.839317      -214.70028819767657
    2024-01-31_10:21:53.840938      -193.20642255951807
    2024-01-31_10:21:56.840750      -181.02017659616354

    Note that Taurus4 did not support exporting if there several attributes in
    a single file, but instead wrote to several files called set001.dat, ...
    """
    Name = "Taurus 4 compatible ASCII export"
    windows = []

    def __init__(self, item):
        Exporter.__init__(self, item)

        self.index_counter = itertools.count(start=0)
        self.data = []
        self.models = []

    def parameters(self):
        return None

    def fileSaveDialog(self, filter=None, opts=None,
                       mode=QtWidgets.QFileDialog.FileMode.AnyFile):
        # Overriding the default method to allow setting the mode
        if opts is None:
            opts = {}
        self.fileDialog = FileDialog()
        self.fileDialog.setFileMode(mode)
        self.fileDialog.setAcceptMode(
            QtWidgets.QFileDialog.AcceptMode.AcceptSave)
        if filter is not None:
            if isinstance(filter, str):
                self.fileDialog.setNameFilter(filter)
            elif isinstance(filter, list):
                self.fileDialog.setNameFilters(filter)
        global LastExportDirectory
        exportDir = LastExportDirectory
        if exportDir is not None:
            self.fileDialog.setDirectory(exportDir)
        self.fileDialog.show()
        self.fileDialog.opts = opts
        self.fileDialog.fileSelected.connect(self.fileSaveFinished)

    def _exportPlotDataItem(self, plotDataItem) -> None:
        if hasattr(plotDataItem, 'getOriginalDataset'):
            # try to access unmapped, unprocessed data
            cd = plotDataItem.getOriginalDataset()
        else:
            # fall back to earlier access method
            cd = plotDataItem.getData()
        if cd[0] is not None:
            # data found, append it
            self.data.append(cd)

    @staticmethod
    def _format_row(row):
        for i, item in enumerate(row):
            if item and i % 2 == 0:
                yield datetime.fromtimestamp(item).isoformat(sep="_")
                continue
            if isinstance(item, str):
                yield item
                continue
            yield np.format_float_positional(item, precision=10)

    def export(self, fileName=None):
        if not isinstance(self.item, PlotItem):
            raise TypeError("Must have a PlotItem selected for CSV export.")

        if not self.data:
            for item in self.item.items:
                if hasattr(item, "getFullModelName"):
                    self.models.append(item.getFullModelName())
                if hasattr(item, 'implements') and item.implements('plotData'):
                    self._exportPlotDataItem(item)

        single = len(self.data) == 1

        if fileName is None:
            if single:
                self.fileSaveDialog(filter=["*.csv", "*.tsv"])
            else:
                # For multiple attributes, the Taurus 4 behavior is
                # to ask for a directory
                self.fileSaveDialog(
                    mode=QtWidgets.QFileDialog.FileMode.Directory)
        else:
            for i, (model, dataset) in enumerate(zip(self.models, self.data),
                                                 start=1):
                if single:
                    path = fileName
                else:
                    path = os.path.join(fileName, f"set{i:03}.dat")
                print("Saving {} at {}".format(model, path))
                with open(path, 'w', newline='') as csvfile:
                    # Write "Taurus 4 like" header
                    # TODO what does the [0] mean here?
                    csvfile.write(f'# DATASET= "{model}[0]"\n')
                    snapshot_time = datetime.now().isoformat(sep="_")
                    csvfile.write(f'# SNAPSHOT_TIME= {snapshot_time}\n')

                    # Write data rows
                    delimiter = "\t"
                    writer = csv.writer(csvfile, delimiter=delimiter,
                                        quoting=csv.QUOTE_MINIMAL)
                    for row in itertools.zip_longest(*dataset, fillvalue=""):
                        writer.writerow(self._format_row(row))

            self.data.clear()


Taurus4TextExporter.register()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################
__all__ = ["PlotLegendTool"]

from taurus.external.qt import QtGui
from taurus.qt.qtcore.configuration import BaseConfigurableClass
import pyqtgraph as pg


class PlotLegendTool(QtGui.QWidgetAction, BaseConfigurableClass):
    """
    This tool adds a legend to the PlotItem to which it is attached, and it
    inserts a checkable menu action for showing/hiding the legend.

    Implementation note: this is implemented as a QWidgetAction+QCheckBox
    instead of a checkable QAction to avoid closing the menu when toggling it
    """

    def __init__(self, parent=None, **kwargs):
        BaseConfigurableClass.__init__(self)
        QtGui.QWidgetAction.__init__(self, parent)
        self._cb = QtGui.QCheckBox()
        self._cb.setText("Show legend")
        self.setDefaultWidget(self._cb)
        self.registerConfigProperty(
            self._cb.isChecked, self._cb.setChecked, "checked"
        )
        # TODO: register config prop for legend position
        self._cb.toggled.connect(self._onToggled)
        self._legend = None
        self._legend_kwargs = kwargs

    def attachToPlotItem(self, plotItem):
        """
        Use this method to add this tool to a plot

        :param plot_item: (PlotItem)
        """
        self._legend = plotItem.addLegend(**self._legend_kwargs)

        # if no explicit pen / brush is set, use some nicer defaults
        if "pen" not in self._legend_kwargs:
            self._legend.setPen(pg.CONFIG_OPTIONS["foreground"])
        if "brush" not in self._legend_kwargs:
            bcolor = pg.mkColor(pg.CONFIG_OPTIONS["background"])
            bcolor.setAlphaF(0.85)
            self._legend.setBrush(bcolor)

        self._cb.setChecked(True)
        menu = plotItem.getViewBox().menu
        menu.addAction(self)

    def _onToggled(self, checked):
        if checked:
            self._legend.show()
        else:
            self._legend.hide()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################


__all__ = ["XAutoPanTool"]

from taurus.external.qt import QtGui, QtCore
from taurus.external.qt import QtWidgets


class XAutoPanTool(QtGui.QAction):
    """
    A tool that provides the "AutoPan" for the X axis of a plot
    (aka "oscilloscope mode"). It is implemented as an Action, and provides a
    method to attach it to a :class:`pyqtgraph.PlotItem`
    """

    def __init__(self, parent=None):
        QtGui.QAction.__init__(self, "Fixed range scale", parent)
        self.setCheckable(True)
        self.toggled.connect(self._onToggled)
        self._timer = QtCore.QTimer()
        self._timer.timeout.connect(self.updateRange)
        self._originalXAutoRange = None
        self._viewBox = None
        self._plotItem = None
        self._XactionMenu = None
        self._scrollStep = 0.2

    def attachToPlotItem(self, plot_item):
        """Use this method to add this tool to a plot

        :param plot_item: (PlotItem)
        """
        self._plotItem = plot_item
        self._viewBox = plot_item.getViewBox()
        self._addToMenu(self._viewBox.menu)
        self._originalXAutoRange = self._viewBox.autoRangeEnabled()[0]
        self._viewBox.sigXRangeChanged.connect(self._onXRangeChanged)
        self._plotItem.fixed_range_checkbox = self

    def _addToMenu(self, menu):
        # In newer PyQtGraph versions (>0.13.1) there is no "menu.axes"
        # property, yet in older versions the QMenu is not added in
        # as child of the menu.
        for submenu in menu.findChildren(QtWidgets.QMenu):
            if submenu.title() == "X axis":
                x_menu = submenu
                break
        else:
            x_menu, _ = menu.axes

        self._plotItem.x_axis_menu = (self._plotItem.getViewBox().menu.ctrl[0]
                                      .gridLayout)
        self._XactionMenu = x_menu.actions()[0]
        x_menu.insertAction(self._XactionMenu, self)
        self.setParent(x_menu)

    def _onToggled(self, checked):
        if checked:
            self._originalXAutoRange = self._viewBox.autoRangeEnabled()[0]
            self._viewBox.enableAutoRange(x=False)

            axisXrange = self._viewBox.state["viewRange"][0]
            x_range = axisXrange[1] - axisXrange[0]

            t = int(x_range / 10.0) * 1000
            t = min(3000, t)
            t = max(50, t)
            self._timer.start(t)
        else:
            self._timer.stop()
            self._viewBox.enableAutoRange(x=self._originalXAutoRange)

        # Get the custom context menu action widget. (X axis)
        # Then disable editing children except for
        # "range_label" and "range_combo_box"
        for element in self.parent().children()[1].children():
            if element.objectName() not in ["range_label", "range_combo_box"]:
                element.setEnabled(not checked)

    def _onXRangeChanged(self):
        self.setChecked(False)

    def updateRange(self):
        """Pans the x axis (change the viewbox range maintaining width but
        ensuring that the right-most point is shown
        """
        dataItems = self._plotItem.listDataItems()
        vbs = set()
        for item in dataItems:
            vbs.add(item.getViewBox())

        if len(dataItems) < 1:
            self._timer.stop()

        # find the largest X axis bound among all viewboxes
        boundMax = float("-inf")
        for vb in vbs:
            cb = vb.childrenBounds()
            if cb is not None and cb[0] is not None and cb[0][1] > boundMax:
                boundMax = cb[0][1]

        if boundMax == float("-inf"):
            return

        axis_X_range, _ = self._viewBox.state["viewRange"]

        x_range = axis_X_range[1] - axis_X_range[0]

        if boundMax > axis_X_range[1] or boundMax < axis_X_range[0]:
            x_min = boundMax - axis_X_range[1]
            x_max = boundMax - axis_X_range[0]
            step = min(max(x_range * self._scrollStep, x_min), x_max)

            self._viewBox.sigXRangeChanged.disconnect(self._onXRangeChanged)
            self._viewBox.setXRange(
                axis_X_range[0] + step,
                axis_X_range[1] + step,
                padding=0.0,
                update=False,
            )
            self._viewBox.sigXRangeChanged.connect(self._onXRangeChanged)



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################
from __future__ import absolute_import

__all__ = ["TaurusTrend"]

import copy
import traceback
from time import localtime, strftime

import numpy as np
from pyqtgraph import PlotWidget
from taurus.core.util.containers import LoopList
from taurus.core.util.log import Logger
from taurus.external.qt import QtGui, PYSIDE2, QtWidgets
from taurus.qt.qtcore.configuration import BaseConfigurableClass

from taurus_pyqtgraph.statisticstool import StatisticsTool
from .autopantool import XAutoPanTool
from .buffersizetool import BufferSizeTool
from .curveproperties import CURVE_COLORS, RangeOptions, set_range_on_trend
from .curvespropertiestool import CurvesPropertiesTool
from .datainspectortool import DataInspectorTool
from .dateaxisitem import DateAxisItem
from .forcedreadtool import ForcedReadTool
from .legendtool import PlotLegendTool
from .taurusmodelchoosertool import TaurusXYModelChooserTool
from .taurustrendset import TaurusTrendSet
from .y2axis import Y2ViewBox
from .titlepatterneditor import TitlePatternEditor, EVALUATION_KEYS

try:
    from pyhdbpp import get_default_reader

    archiving_reader = get_default_reader()
except Exception:
    archiving_reader = None

SECONDS_48_HOURS = 172800


class TaurusTrend(PlotWidget, BaseConfigurableClass):
    """
    TaurusTrend is a general widget for plotting the evolution of a value
    over time. It is an extended taurus-aware version of
    :class:`pyqtgraph.PlotWidget`.

    Apart from all the features already available in a regulat PlotWidget,
    TaurusTrend incorporates the following tools/features:

        - Secondary Y axis (right axis)
        - Time X axis
        - A plot configuration dialog, and save/restore configuration
          facilities
        - A menu option for adding/removing taurus  models
        - A menu option for showing/hiding the legend
        - Automatic color change of curves for newly added models

    """

    def __init__(self, parent=None, **kwargs):

        buffer_size = kwargs.pop("buffer_size", 65536)

        if PYSIDE2:
            # Workaround for https://bugreports.qt.io/browse/PYSIDE-1564
            BaseConfigurableClass.__init__(self)
            PlotWidget.__init__(self, parent=parent, **kwargs)
        else:
            super(TaurusTrend, self).__init__(parent=parent, **kwargs)

        # Compose with a Logger
        self._logger = Logger(name=self.__class__.__name__)
        self.debug = self._logger.debug
        self.info = self._logger.info
        self.warning = self._logger.warning
        self.error = self._logger.error

        # set up cyclic color generator
        self._curveColors = LoopList(CURVE_COLORS)
        self._curveColors.setCurrentIndex(-1)

        self.titlePatternEditor = TitlePatternEditor()

        plot_item = self.getPlotItem()
        menu = plot_item.getViewBox().menu

        # add trends clear action
        clearTrendsAction = QtGui.QAction("Clear trends", menu)
        clearTrendsAction.triggered.connect(self.clearTrends)
        menu.addAction(clearTrendsAction)

        # add save & retrieve configuration actions
        saveConfigAction = QtGui.QAction("Save configuration", menu)
        saveConfigAction.triggered.connect(self._onSaveConfigAction)
        menu.addAction(saveConfigAction)

        loadConfigAction = QtGui.QAction("Retrieve saved configuration", menu)
        loadConfigAction.triggered.connect(self._onLoadConfigAction)
        menu.addAction(loadConfigAction)

        # add change curve labels
        changeCurvesTitlesAction = QtGui.QAction("Change Curves Titles "
                                                 "(All curves)...", menu)
        changeCurvesTitlesAction.triggered.connect(
            self._onChangeCurvesTitlesAction)
        menu.addAction(changeCurvesTitlesAction)

        # set up archiving functionality
        self._archiving_enabled = False
        self._archiving_reader = None
        self._auto_reload_checkbox = None
        self._dismiss_archive_message = False
        if self._setArchivingReader():
            self._loadArchivingContextActions()

        self.registerConfigProperty(self._getState, self.restoreState, "state")

        # add legend tool
        legend_tool = PlotLegendTool(self)
        legend_tool.attachToPlotItem(plot_item)

        # add model chooser
        self._model_chooser_tool = TaurusXYModelChooserTool(
            self, itemClass=TaurusTrendSet, showX=False
        )
        self._model_chooser_tool.attachToPlotItem(
            self.getPlotItem(), self, self._curveColors
        )

        # add statistics tool
        self._statistics_tool = StatisticsTool(self, itemClass=TaurusTrendSet)
        self._statistics_tool.attachToPlotItem(self.getPlotItem(), self)

        # add Y2 axis
        self._y2 = Y2ViewBox()
        self._y2.attachToPlotItem(plot_item)

        # Add time X axis
        axis = DateAxisItem(orientation="bottom")
        axis.attachToPlotItem(plot_item)

        # add plot configuration dialog
        self._cprop_tool = CurvesPropertiesTool(self)
        self._cprop_tool.attachToPlotItem(plot_item, y2=self._y2)

        # add data inspector widget
        inspector_tool = DataInspectorTool(self)
        inspector_tool.attachToPlotItem(self.getPlotItem())

        # add force read tool
        self._fr_tool = ForcedReadTool(self)
        self._fr_tool.attachToPlotItem(self.getPlotItem())

        # add buffer size tool
        self.buffer_tool = BufferSizeTool(self, buffer_size=buffer_size)
        self.buffer_tool.attachToPlotItem(self.getPlotItem())

        # Add the auto-pan ("oscilloscope mode") tool
        self._autopan = XAutoPanTool()
        self._autopan.attachToPlotItem(self.getPlotItem())

        # add Set View Range actions
        x_axis_menu = self.plotItem.x_axis_menu

        label = QtGui.QLabel("Set View Range: ")
        label.setObjectName("range_label")
        self.combo_box = QtGui.QComboBox(self)
        self.combo_box.setObjectName("range_combo_box")
        self.combo_box.setEditable(True)
        for option in RangeOptions:
            self.combo_box.addItem(str(option))

        x_axis_menu.addWidget(label, 8, 0, 1, 2)
        x_axis_menu.addWidget(self.combo_box, 8, 2, 1, 2)

        self.combo_box.activated.connect(self._on_range_idx_changed)

        # Register config properties
        self.registerConfigDelegate(self._model_chooser_tool, "XYmodelchooser")
        # self.registerConfigDelegate(self._y2, "Y2Axis")
        self.registerConfigDelegate(self._cprop_tool, "CurvePropertiesTool")
        self.registerConfigDelegate(legend_tool, "legend")
        self.registerConfigDelegate(self._fr_tool, "forceread")
        self.registerConfigDelegate(self.buffer_tool, "buffer")
        self.registerConfigDelegate(inspector_tool, "inspector")

    def __getitem__(self, idx):
        """
        Provides a list-like interface: items can be accessed using slice
        notation
        """
        return self._getCurves()[idx]

    def __len__(self):
        return len(self._getCurves())

    def __bool__(self):
        return True

    def _loadArchivingContextActions(self):
        """Loads archiving options to context menu on the trend (right-click)
        and enables triggers regarding archiving.
        """
        menu = self.plotItem.getViewBox().menu

        archiving_menu = QtGui.QMenu("Archiving", menu)
        menu.addMenu(archiving_menu)

        self._auto_reload_checkbox = QtGui.QAction(
            "Autoreload", archiving_menu
        )
        self._auto_reload_checkbox.setCheckable(True)
        self._auto_reload_checkbox.setChecked(False)
        self._auto_reload_checkbox.triggered.connect(
            self._onEnableDisableArchivingClicked
        )

        load_once_action = QtGui.QAction("Load Once (Ctrl+L)", archiving_menu)
        load_once_action.triggered.connect(self._loadArchivingDataOnce)

        load_once_shortcut = QtGui.QShortcut(QtGui.QKeySequence("Ctrl+L"),
                                             self)
        load_once_shortcut.activated.connect(self._loadArchivingDataOnce)

        decimate_and_redraw = QtGui.QAction("Decimate and Redraw (Ctrl+R)",
                                            archiving_menu)
        decimate_and_redraw.triggered.connect(self._decimate_and_redraw)

        decimate_shortcut = QtGui.QShortcut(QtGui.QKeySequence("Ctrl+R"), self)
        decimate_shortcut.activated.connect(self._decimate_and_redraw)

        archiving_menu.addAction(self._auto_reload_checkbox)
        archiving_menu.addAction(load_once_action)
        archiving_menu.addAction(decimate_and_redraw)

    def _decimate_and_redraw(self):
        self.clearTrends()
        self._loadArchivingDataOnce()

    def _loadArchivingDataOnce(self):
        self._loadArchivingData(loadOnce=True)

    def _onEnableDisableArchivingClicked(self):
        """Change the state of boolean archiving_enabled to the opposite
        of their actual value. If it's set to True a connection between
        sigRangeChanged and local function on_changed_trend is set,
        otherwise, the connection gets disconnected.
        """
        self._archiving_enabled = not self._archiving_enabled

        if self._archiving_enabled:
            self.sigRangeChanged \
                .connect(lambda: self._loadArchivingData(loadOnce=False))
            self.info("Archiving option set to enabled")
            self._loadArchivingData(loadOnce=False)  # Force first data query
        else:
            self.sigRangeChanged.disconnect()
            self.info("Archiving option set to disabled")

    def _setArchivingReader(self):
        """Try to set up a reader and return if it was possible or not
        (True/False). :return: True if reader is set or False if not
        """
        if archiving_reader:
            self._archiving_reader = archiving_reader
            self.info("Archiving reader set")
            return True
        else:
            self.info("Archiving reader not set")
            return False

    def _loadArchivingData(self, loadOnce=False):
        """When there is a change on the Range of view perform
        a query to get archiving data and append it to the left.
        """

        for taurus_trend_set in self.plotItem.dataItems:
            if not isinstance(taurus_trend_set, TaurusTrendSet):
                continue
            try:
                left_time_range_value = self.visibleRange().left()
                right_time_range_value = self.visibleRange().right()
                plot_time_range = \
                    right_time_range_value - left_time_range_value
                elder_buffer_value = (taurus_trend_set._xBuffer[0]
                                      if len(
                    taurus_trend_set._xBuffer) else right_time_range_value)
                query_window = elder_buffer_value - left_time_range_value

                if query_window > (.15 * plot_time_range) > 1:

                    from_date = strftime('%Y-%m-%dT%H:%M:%S',
                                         localtime(left_time_range_value))
                    to_date = strftime('%Y-%m-%dT%H:%M:%S',
                                       localtime(elder_buffer_value))
                    if self._checkForQuerySizeAndUserConfirmation(
                        left_time_range_value, elder_buffer_value,
                            taurus_trend_set.modelName):
                        try:
                            values = self._archiving_reader \
                                .get_attribute_values(
                                    taurus_trend_set.modelName,
                                    from_date, to_date, decimate=True)
                            self.debug(
                                'loadArchivingData(%s, %s, %s): %d values' % (
                                    taurus_trend_set.modelName, from_date,
                                    to_date, len(values)))

                        except KeyError as ke:
                            values = None
                            self.debug(
                                "Attribute '{}' has no archiving data".format(
                                    ke))

                        if values is not None and len(values):
                            if (len(values) + len(taurus_trend_set._xBuffer)) \
                                    < self.buffer_tool.bufferSize():
                                self.debug(
                                    "left-appending historical data from {} "
                                    "to {}".format(from_date, to_date))
                                x, y, _, _ = zip(*values)
                                x = np.array(x)
                                y = np.array(y)
                                y.shape = (len(y), 1)
                                try:
                                    taurus_trend_set._xBuffer.extendLeft(x)
                                    taurus_trend_set._yBuffer.extendLeft(y)
                                    taurus_trend_set._update()
                                except ValueError as e:
                                    self.error("Error left-appending data to "
                                               "buffer.\n", e)
                            else:
                                msg = ("Buffer size is surpassing limit and "
                                       "data has been discarded.\n "
                                       "You can change the buffer size at "
                                       "your "
                                       "own responsibility and try again.")
                                if loadOnce:
                                    self._askForConfirmation(msg, buttons=QtGui
                                                             .QMessageBox.Ok)
                                else:
                                    msg += "\nAuto reload has been disabled"
                                    self._disableAutoReloadAndDiscardData(msg)
                                    return
                    else:
                        if not loadOnce:
                            msg = "Data from archiving has been discarded " \
                                  "and reload disabled"
                            self._disableAutoReloadAndDiscardData(msg)
                        break
            except Exception as e:
                self.warning("Error updating trend set of model {} "
                             "with error {}".format(taurus_trend_set.modelName
                                                    , e))

    def _disableAutoReloadAndDiscardData(self, message):
        self._askForConfirmation(message, buttons=QtGui.QMessageBox.Ok)
        self.info(message)
        self._auto_reload_checkbox.setChecked(False)
        self._onEnableDisableArchivingClicked()  # Force a trigger

    def _on_range_idx_changed(self):
        time_range = self.combo_box.currentText()
        plot_item = self.getPlotItem()
        set_range_on_trend(time_range, plotItem=plot_item)
        plot_item.getViewBox().menu.close()

    def _getCurves(self):
        """returns a flat list with all items from all trend sets"""
        ret = []
        for ts in self.getTrendSets():
            ret += ts[:]
        return ret

    def getTrendSets(self):
        """Returns all the trend sets attached to this plot item"""
        return [
            e
            for e in self.getPlotItem().listDataItems()
            if isinstance(e, TaurusTrendSet)
        ]

    def clearTrends(self):
        """Clear the buffers of all the trend sets in the plot"""
        for ts in self.getTrendSets():
            ts.clearBuffer()

    def setModel(self, names):
        """Set a list of models"""
        # support passing a string  in names instead of a sequence
        if isinstance(names, str):
            names = [names]
        self._model_chooser_tool.updateModels(names or [])

    def addModels(self, names):
        """Reimplemented to delegate to the  model chooser"""
        # support passing a string in names
        if isinstance(names, str):
            names = [names]
        self._model_chooser_tool.addModels(names)

    def _getState(self):
        """Same as PlotWidget.saveState but removing viewRange conf to force
        a refresh with targetRange when loading
        """
        state = copy.deepcopy(self.saveState())
        # remove viewRange conf
        del state["view"]["viewRange"]
        return state

    def setXAxisMode(self, x_axis_mode):
        """Required generic TaurusTrend API """
        if x_axis_mode != "t":
            raise NotImplementedError(  # TODO
                'X mode "{}" not yet supported'.format(x_axis_mode)
            )

    def setAxisLogMode(self, axis, isEnabled=True):
        """
        Method to change log mode of a desired axis.
        Possible options:
            setAxisLogMode("left", True)
            setAxisLogMode("bottom", True)
            setAxisLogMode("y", True) # The same as left
            setAxisLogMode("x", True) # The same as bottom
        """

        if axis in ["left", "y"]:
            self.plotItem.getAxis("left").setLogMode(y=isEnabled)
            axis = "left"
        elif axis in ["bottom", "x"]:
            self.plotItem.getAxis("bottom").setLogMode(x=isEnabled)
            axis = "bottom"
        else:
            axis = False

        if axis:
            menu = self.plotItem.getViewBox().menu

            if hasattr(menu, "axes"):
                x_menu, y_menu = menu.axes
                actions = []
                if axis == "left":
                    actions = y_menu.actions()
                elif axis == "bottom":
                    actions = x_menu.actions()

                for action in actions:
                    if action.text() == "Log scale":
                        action.setChecked(isEnabled)
            else:
                submenus = menu.findChildren(QtWidgets.QMenu)
                menu_title = "Y axis" if axis == "left" else "X axis"
                for submenu in submenus:
                    if submenu.title() == menu_title:
                        for action in submenu.actions():
                            if action.text() == "Log scale":
                                action.setChecked(isEnabled)

    def setForcedReadingPeriod(self, period):
        """Required generic TaurusTrend API """
        self._fr_tool.setPeriod(period)

    def setMaxDataBufferSize(self, buffer_size):
        """Required generic TaurusTrend API """
        self.buffer_tool.setBufferSize(buffer_size)

    def _onSaveConfigAction(self):
        """wrapper to avoid issues with overloaded signals"""
        return self.saveConfigFile()

    def _onLoadConfigAction(self):
        """wrapper to avoid issues with overloaded signals"""
        return self.loadConfigFile()

    def _onChangeCurvesTitlesAction(self):
        """Show pop up to change title on curves"""
        label_config = self.titlePatternEditor.showDialog()

        if label_config:
            for item in self.getTrendSets():
                result = label_config
                for key, resolver in EVALUATION_KEYS.items():
                    if key in self.titlePatternEditor.legend_pattern:
                        try:
                            result = result.replace(key,
                                                    resolver(item.modelObj))
                        except AttributeError:
                            self._logger.warning(traceback.format_exc())
                            result = result.replace(key, "NoDef")

                if len(item) == 1:
                    item[0].setData(name=result)
                    self.plotItem.legend.getLabel(item[0]).setText(result)

                if len(item) > 1:
                    for i in range(len(item)):
                        result_array = "{}[{}]".format(result, i)
                        item[i].setData(name=result_array)
                        (self.plotItem.legend.getLabel(item[i])
                         .setText(result_array))

    def _checkForQuerySizeAndUserConfirmation(self, from_date, to_date,
                                              model_name):
        if to_date - from_date > SECONDS_48_HOURS:
            if not self._dismiss_archive_message:
                buttonClicked = self._askForConfirmation(
                    "You are querying more than 48 hours for "
                    "{}, this may take a while.\nContinue? (Yes to All "
                    "disables "
                    "this message) ".format(model_name))
                self._dismiss_archive_message = \
                    buttonClicked == QtGui.QMessageBox.YesToAll
                return buttonClicked in [QtGui.QMessageBox.Ok,
                                         QtGui.QMessageBox.YesToAll]

        return True

    def _askForConfirmation(self, message,
                            buttons=QtGui.QMessageBox.Ok | QtGui.QMessageBox
                            .YesToAll | QtGui.QMessageBox.Cancel):
        warn_user = QtGui.QMessageBox(QtGui.QMessageBox.Warning, "Warning!",
                                      message, buttons)
        return warn_user.exec_()


def trend_main(
    models=(), config_file=None, demo=False, window_name="TaurusTrend (pg)"
):
    """Launch a TaurusTrend"""
    import sys
    from taurus.qt.qtgui.application import TaurusApplication

    app = TaurusApplication(cmd_line_parser=None, app_name="taurustrend(pg)")

    w = TaurusTrend()

    w.setWindowTitle(window_name)

    if demo:
        models = list(models)
        models.extend(["eval:rand()", "eval:1+rand(2)"])

    if config_file is not None:
        w.loadConfigFile(config_file)

    if models:
        w.setModel(models)

    w.show()
    ret = app.exec_()

    sys.exit(ret)


if __name__ == "__main__":
    trend_main(models=("eval:rand()", "sys/tg_test/1/ampli"),)



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################
__all__ = ["ForcedReadTool"]

from taurus.external.qt import QtGui, QtCore
from taurus.qt.qtcore.configuration import BaseConfigurableClass


class ForcedReadTool(QtGui.QAction, BaseConfigurableClass):
    """
    This tool provides a menu option to control the "Forced Read" period of
    Plot data items that implement a `setForcedReadPeriod` method
    (see, e.g. :meth:`TaurusTrendSet.setForcedReadPeriod`).

    The force-read feature consists on forcing periodic attribute reads for
    those attributes being plotted with a :class:`TaurusTrendSet` object.
    This allows to force plotting periodical updates even for attributes
    for which the taurus polling is not enabled.
    Note that this is done at the widget level and therefore does not affect
    the rate of arrival of events for other widgets connected to the same
    attributes

    This tool inserts an action with a spinbox and emits a `valueChanged`
    signal whenever the value is changed.

    The connection between the data items and this tool can be done manually
    (by connecting to the `valueChanged` signal or automatically, if
    :meth:`autoconnect()` is `True` (default). The autoconnection feature works
    by discovering the compliant data items that share associated to the
    plot_item.

    This tool is implemented as an Action, and provides a method to attach it
    to a :class:`pyqtgraph.PlotItem`
    """

    valueChanged = QtCore.pyqtSignal(int)

    def __init__(
        self,
        parent=None,
        period=0,
        text="Change forced read period...",
        autoconnect=True,
    ):
        BaseConfigurableClass.__init__(self)
        QtGui.QAction.__init__(self, text, parent)
        tt = "Period between forced readings.\nSet to 0 to disable"
        self.setToolTip(tt)
        self._period = period
        self._autoconnect = autoconnect

        # register config properties
        self.registerConfigProperty(self.period, self.setPeriod, "period")
        self.registerConfigProperty(
            self.autoconnect, self.setAutoconnect, "autoconnect"
        )

        # internal conections
        self.triggered.connect(self._onTriggered)

    def _onTriggered(self):
        period = self.period()
        period, ok = QtGui.QInputDialog.getInt(
            self.parentWidget(),
            "New read period",
            "Period (in ms) between forced readings.\nSet to 0 to disable",
            period,
            0,
            604800000,
            500,
        )
        if ok:
            self.setPeriod(period)

    def attachToPlotItem(self, plot_item):
        """Use this method to add this tool to a plot

        :param plot_item: (PlotItem)
        """
        menu = plot_item.getViewBox().menu
        menu.addAction(self)
        self.plot_item = plot_item
        # force an update of period for connected trendsets
        self.setPeriod(self.period())
        if self.autoconnect():
            # enable the forced reading also for trendsets added in the future
            try:  # requires https://github.com/pyqtgraph/pyqtgraph/pull/1388
                plot_item.scene().sigItemAdded.connect(self._onAddedItem)
            except AttributeError:
                pass

    def _onAddedItem(self, item):
        if hasattr(item, "setForcedReadPeriod"):
            item.setForcedReadPeriod(self.period())

    def autoconnect(self):
        """Returns autoconnect state

        :return: (bool)
        """
        return self._autoconnect

    def setAutoconnect(self, autoconnect):
        """Set autoconnect state. If True, the tool will autodetect trendsets
        associated to the plot item and will call setForcedReadPeriod
        on each of them for each change. If False, it will only emit a
        valueChanged signal and only those connected to it will be notified
        of changes

        :param autoconnect: (bool)
        """
        self._autoconnect = autoconnect

    def period(self):
        """Returns the current period value (in ms)

        :return: (int)
        """
        return self._period

    def setPeriod(self, period):
        """Change the period value. Use 0 for disabling

        :param period: (int) period in ms
        """
        self._period = period
        # update existing items
        if self.autoconnect() and self.plot_item is not None:
            for item in self.plot_item.listDataItems():
                if hasattr(item, "setForcedReadPeriod"):
                    item.setForcedReadPeriod(period)
        # emit valueChanged
        self.valueChanged.emit(period)


if __name__ == "__main__":
    import taurus

    taurus.setLogLevel(taurus.Debug)
    import sys
    from taurus.qt.qtgui.application import TaurusApplication
    from taurus_pyqtgraph import TaurusTrendSet, DateAxisItem
    import pyqtgraph as pg

    # from taurus_pyqtgraph import ForcedReadTool

    app = TaurusApplication()

    w = pg.PlotWidget()

    axis = DateAxisItem(orientation="bottom")
    w = pg.PlotWidget()
    axis.attachToPlotItem(w.getPlotItem())

    # test adding the curve before the tool
    ts1 = TaurusTrendSet(name="before", symbol="o")
    ts1.setModel("eval:rand()+1")

    w.addItem(ts1)

    fr = ForcedReadTool(w, period=1000)
    fr.attachToPlotItem(w.getPlotItem())

    # test adding the curve after the tool
    ts2 = TaurusTrendSet(name="after", symbol="+")
    ts2.setModel("eval:rand()")

    w.addItem(ts2)

    w.show()

    ret = app.exec_()

    import pprint

    pprint.pprint(fr.createConfig())

    sys.exit(ret)



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################


"""
Example on how to use a separate widget (LegendItem) for the legend of a plot.
(Pure Qt)
"""

import pyqtgraph as pg
from pyqtgraph.Qt import QtGui
import sys


if __name__ == "__main__":

    app = QtGui.QApplication([])

    # instantiate the main plot
    plt = pg.PlotWidget()
    plt.setWindowTitle("pyqtgraph example: PLOT")

    # instantiate a graphics view to contain the legend
    gv = QtGui.QGraphicsView(QtGui.QGraphicsScene())
    gv.setWindowTitle("pyqtgraph example: Legend")
    gv.setBackgroundBrush(QtGui.QBrush(QtGui.QColor("black")))

    legend = pg.LegendItem(size=(100, 60), offset=(70, 30))
    gv.scene().addItem(legend)

    # create 3 curves
    c1 = plt.plot(
        [1, 3, 2, 4],
        pen="r",
        symbol="o",
        symbolPen="r",
        symbolBrush=0.5,
        name="red plot",
    )
    c2 = plt.plot(
        [2, 1, 4, 3],
        pen="g",
        fillLevel=0,
        fillBrush=(255, 255, 255, 30),
        name="green plot",
    )
    c3 = plt.plot(list(range(7)), pen="c", fillLevel=0)

    # add the **named** curves to the legend
    for dataitem in plt.getPlotItem().listDataItems():
        if dataitem.name():
            legend.addItem(dataitem, dataitem.name())

    plt.show()
    gv.show()

    sys.exit(app.exec_())



import numpy

import taurus_pyqtgraph as tpg
import pyqtgraph as pg
from collections import Counter
import time
from .util import show_and_wait, get_sub_config  # noqa


# def test_trend_teardown(qtbot):
#     """
#     For some reason (to be investigated) the teardown of this tests
#     triggers a crash in pytest (maybe related to qt?)
#     """
#     w = tpg.TaurusTrend()
#     qtbot.addWidget(w)
#     m = ["eval:1", "eval:2"]
#     # m = ["eval:1", "eval:2"]  # uncommenting this line avoids the trigger
#     # m = ["eval:1", "eval:2"]  # uncommenting this line avoids the trigger
#     w.setModel(m)
#     # w.setModel([])  # uncommenting this line avoids the trigger


def test_trend_model_setting(qtbot):

    w = tpg.TaurusTrend()
    qtbot.addWidget(w)

    vb1 = w.getViewBox()
    vb2 = w._y2

    assert len(w) == 0
    assert w[:] == []
    assert w._model_chooser_tool.getModelNames() == []
    assert vb1.addedItems == []
    assert vb2.addedItems == []

    models = [
        "eval:1+rand()",
        "eval:2+rand(2)",
        "eval:3+rand(3)",
    ]

    w.setModel(models)
    sets = w.getTrendSets()
    ts0 = sets[0]
    ts1 = sets[1]
    ts2 = sets[2]

    assert len(w.getTrendSets()) == 3
    assert len(w) == 6
    assert len(ts0) == 1
    assert len(ts1) == 2
    assert len(ts2) == 3
    assert len(w._model_chooser_tool._getCurveInfo()) == 3

    for m, (x, y, n) in zip(models, w._model_chooser_tool._getCurveInfo()):
        assert x is None
        assert y.endswith(m[5:])
        assert isinstance(n, list)
        assert len(n) > 1
        for cn in n:
            assert "+rand(" in cn
        for cn in n[1:]:
            assert cn[-1] == "]"

    assert ts0[0].name() == "1+rand()[0]"
    assert ts1[0].name() == "2+rand(2)[0]"
    assert ts1[1].name() == "2+rand(2)[1]"
    assert ts2[0].name() == "3+rand(3)[0]"
    assert ts2[1].name() == "3+rand(3)[1]"
    assert ts2[2].name() == "3+rand(3)[2]"

    assert w._model_chooser_tool.getModelNames() == [
        ts.getFullModelNames() for ts in w.getTrendSets()
    ]

    # check items of viewboxes (for vb1, ignore the sorting)
    assert Counter(vb1.addedItems) == Counter(sets + ts0[:] + ts1[:] + ts2[:])
    assert vb2.addedItems == []

    # move the whole ts1 to Y2
    # vb1.removeItem(ts1)
    # vb2.addItem(ts1)  # TODO: NOT WORKING (crashes!)
    # assert Counter(vb1.addedItems) == Counter([ts0, ts2] + ts0[:] + ts2[:])
    # assert Counter(vb1.addedItems) == Counter([ts1] + ts1[:])

    # move first curve of each trendset to Y2
    for ts in sets:
        c = ts[0]
        vb1.removeItem(c)
        vb2.addItem(c)
    assert Counter(vb1.addedItems) == Counter(sets + ts1[1:] + ts2[1:])
    assert vb2.addedItems == [ts0[0], ts1[0], ts2[0]]

    # check that there are no duplications because of moving
    assert len(w.getTrendSets()) == 3
    assert len(w) == 6
    assert len(ts0) == 1
    assert len(ts1) == 2
    assert len(ts2) == 3
    assert len(w._model_chooser_tool._getCurveInfo()) == 3

    # add a regular data item (non-taurus) to y2
    c0 = pg.PlotDataItem(name="pg item", pen="y", fillLevel=0, brush="y")
    now = time.time()
    c0.setData(numpy.linspace(now - 5, now + 5, 10), 1 - numpy.random.rand(10))
    vb2.addItem(c0)

    # check that the regular item is not counted as a trendset item...
    assert w.getTrendSets() == [ts0, ts1, ts2]
    assert w._model_chooser_tool.getModelNames() == [
        ts.getFullModelNames() for ts in w.getTrendSets()
    ]
    # ... but still, it is added to y2
    assert Counter(vb1.addedItems) == Counter(sets + ts1[1:] + ts2[1:])
    assert vb2.addedItems == [ts0[0], ts1[0], ts2[0], c0]

    # manually add a trendset to y1 and move its first curve to y2
    ts3 = tpg.TaurusTrendSet(name="TS-3", symbol="o")
    ts3.setModel('eval:Quantity(rand(2)-1,"m")')
    w.addItem(ts3)
    vb1.removeItem(ts3[0])
    vb2.addItem(ts3[0])

    assert len(ts3) == 2
    sets = w.getTrendSets()
    assert sets == [ts0, ts1, ts2, ts3]
    assert len(w) == 8
    assert w[:] == ts0[:] + ts1[:] + ts2[:] + ts3[:]
    assert Counter(vb1.addedItems) == Counter(
        sets + ts1[1:] + ts2[1:] + ts3[1:]
    )
    assert Counter(vb2.addedItems) == Counter(
        [ts0[0], ts1[0], ts2[0], c0, ts3[0]]
    )
    assert w._model_chooser_tool.getModelNames() == [
        ts0.getFullModelNames(),
        ts1.getFullModelNames(),
        ts2.getFullModelNames(),
        ts3.getFullModelNames(),
    ]

    # manually add a trendset to y1
    ts4 = tpg.TaurusTrendSet(name="TS-4", symbol="s")
    ts4.setModel('eval:Quantity(rand()-2,"m")')
    w.addItem(ts4)

    assert len(ts4) == 1
    sets = w.getTrendSets()
    assert sets == [ts0, ts1, ts2, ts3, ts4]
    assert Counter(vb1.addedItems) == Counter(
        sets + ts1[1:] + ts2[1:] + ts3[1:] + ts4[:]
    )
    assert Counter(vb2.addedItems) == Counter(
        [ts0[0], ts1[0], ts2[0], c0, ts3[0]]
    )
    assert w._model_chooser_tool.getModelNames() == [
        ts0.getFullModelNames(),
        ts1.getFullModelNames(),
        ts2.getFullModelNames(),
        ts3.getFullModelNames(),
        ts4.getFullModelNames(),
    ]

    # show_and_wait(qtbot, w)  # uncomment for visually checking

    # Add existing (ts1) model again and check that nothing is recreated
    ts0_0 = ts0[0]
    ts1_0, ts1_1 = ts1[:]
    ts2_0, ts2_1, ts2_2 = ts2[:]
    ts3_0, ts3_1 = ts3[:]
    ts4_0 = ts4[0]

    w.addModels([models[1]])
    sets = w.getTrendSets()

    assert sets == [ts0, ts1, ts2, ts3, ts4]
    assert len(w) == 9
    assert w[:] == [
        ts0_0,
        ts1_0,
        ts1_1,
        ts2_0,
        ts2_1,
        ts2_2,
        ts3_0,
        ts3_1,
        ts4_0,
    ]

    # -----------------------------------------------------------------------
    # nothing is recreated or duplicated, but all the trendsets are reset
    # and their curves restored to the trendset viewbox even if they had
    # been moved to another one.
    # TODO: check if we want to avoid that and have the following instead
    # assert Counter(vb1.addedItems) == Counter(
    #     [ts0, ts1, ts2, ts4] + ts1[1:] + ts2[1:] + ts3[1:] + ts4[:]
    # )
    # assert Counter(vb2.addedItems) == Counter(
    #     [ts0[0], ts1[0], ts2[0], c0, ts3, ts3[0]]
    # )
    assert Counter(vb1.addedItems) == Counter(
        sets + ts0[:] + ts1[:] + ts2[:] + ts3[:] + ts4[:]
    )
    assert vb2.addedItems == [c0]
    # -----------------------------------------------------------------------
    assert w._model_chooser_tool.getModelNames() == [
        ts0.getFullModelNames(),
        ts1.getFullModelNames(),
        ts2.getFullModelNames(),
        ts3.getFullModelNames(),
        ts4.getFullModelNames(),
    ]

    # set (not adding!) 1 model which is already on y1
    # only the non-taurus curve and the just set trendset remain.
    w.setModel([models[1]])
    assert w.getTrendSets() == [ts1]  # ts1 **is** still the same object!
    assert w[:] == [ts1[0], ts1[1]]
    assert Counter(vb1.addedItems) == Counter([ts1] + ts1[:])
    assert vb2.addedItems == [c0]
    assert w._model_chooser_tool.getModelNames() == [ts1.getFullModelNames()]

    # set empty model (not adding!, the non taurus curve is kept)
    w.setModel([])
    assert w.getTrendSets() == []
    assert w[:] == []
    assert vb1.addedItems == []
    assert vb2.addedItems == [c0]
    assert w._model_chooser_tool.getModelNames() == []

    # show_and_wait(qtbot, w)  # uncomment for visually checking

    # remove non-taurus curve
    vb2.removeItem(c0)
    assert w[:] == []
    assert vb1.addedItems == []
    assert vb2.addedItems == []
    assert w._model_chooser_tool.getModelNames() == []

    # show_and_wait(qtbot, w)  # uncomment for visually checking


def test_modelchooser_config(qtbot):

    w1 = tpg.TaurusTrend()
    qtbot.addWidget(w1)

    models1 = [
        "eval:1*rand()",
        (None, "eval:2*rand(2)", "foo"),
    ]

    w1.setModel(models1)

    # test createConfig
    cfg = w1.createConfig()
    xymccfg1 = get_sub_config(cfg, "XYmodelchooser")
    modelscfg1 = get_sub_config(xymccfg1, "CurveInfo")
    assert modelscfg1[0] == (
        None,
        "eval://localhost/@DefaultEvaluator/1*rand()",
        ['1*rand()', '1*rand()[0]'],
    )
    assert modelscfg1[1] == (
        None,
        "eval://localhost/@DefaultEvaluator/2*rand(2)",
        ["foo", "foo[0]", "foo[1]"],
    )

    # test applyConfig
    w2 = tpg.TaurusTrend()
    qtbot.addWidget(w2)
    assert w2.getTrendSets() == []
    assert len(w2) == 0
    assert len(w2._model_chooser_tool.getModelNames()) == 0

    # add a model to w2
    w2.setModel("eval:3*rand()")
    assert len(w2.getTrendSets()) == 1
    assert len(w2) == 1
    assert len(w2._model_chooser_tool.getModelNames()) == 1

    # apply config (the previously added tauruscurve of w2 should be removed)
    w2.applyConfig(cfg)
    assert len(w2.getTrendSets()) == 2
    assert len(w2) == 3
    assert [type(ts) for ts in w2.getTrendSets()] == [tpg.TaurusTrendSet] * 2
    assert [type(c) for c in w2] == [tpg.TrendCurve] * 3
    assert len(w2._model_chooser_tool.getModelNames()) == 2
    assert w2._model_chooser_tool._getCurveInfo() == modelscfg1

    # show_and_wait(qtbot, w1, w2)  # uncomment for visually checking

    # avoid teardown issues
    w1.setModel(None)
    w2.setModel(None)


def test_curveproperties_configfile(qtbot, tmp_path):
    w1 = tpg.TaurusTrend()
    qtbot.addWidget(w1)
    w1.setBackground(0.3)

    w1vb1 = w1.getViewBox()
    w1vb2 = w1._y2

    # add a regular data item (non-taurus) to y1
    c0 = pg.PlotDataItem(name="PG-0", pen="y", fillLevel=0, brush="y")
    now = time.time()
    c0.setData(numpy.linspace(now - 5, now + 5, 10), 1 - numpy.random.rand(10))
    w1.getPlotItem().addItem(c0)

    # add a regular data item (non-taurus) to y2
    c1 = pg.PlotDataItem(name="PG-1", pen="y", symbol="d", symbolBrush="r")
    c1.setData(
        numpy.linspace(now - 5, now + 5, 10),
        1 - numpy.linspace(0, 20, 10) / 20.0,
    )
    w1vb2.addItem(c1)

    # add a TrendSet to y1
    ts0 = tpg.TaurusTrendSet(name="TS0", symbol="o")
    ts0.setModel('eval:Quantity(2+rand(),"m")')
    w1.addItem(ts0)
    c2 = ts0[0]

    # add a TrendSet to y1 and move its curve to Y2
    ts1 = tpg.TaurusTrendSet(name="TS1", symbol="s")
    ts1.setModel('eval:Quantity(3+rand(),"km")')
    w1.addItem(ts1)
    c3 = ts1[0]
    w1vb1.removeItem(c3)
    w1vb2.addItem(c3)
    assert c3 in w1vb2.addedItems

    # Save a config file at this point
    f1 = tmp_path / "trend1.pck"
    with open(str(f1), "wb") as ofile:
        w1.saveConfigFile(ofile=ofile)
    assert f1.exists()
    assert len(list(tmp_path.iterdir())) == 1

    # Add 2 TrendSets using addModel
    models1 = [
        (None, "eval:1*rand(2)", "TS2"),
        (None, "eval:2*rand(2)", "TS3"),
    ]
    w1.addModels(models1)

    # ------------------------------------------------------------------
    # After calling addModels, TrendCurves that were manually moved to
    #  Y2 are reset to Y1 (where their trendset is).
    # TODO: the following 2 lines should not be needed
    assert c3 in w1vb1.addedItems  # this is what it is, not what it should be
    w1vb2.addItem(c3)  # manually move c3 to Y2 again...
    assert c3 not in w1vb1.addedItems
    assert c3 in w1vb2.addedItems
    # ------------------------------------------------------------------

    ts2 = w1.getTrendSets()[2]
    c4, c5 = ts2[:]
    ts3 = w1.getTrendSets()[3]
    c6, c7 = ts3[:]

    c4.setPen("g")

    c5.setPen(None)
    c5.setSymbol("t")
    c5.setSymbolSize(7)
    c5.setSymbolBrush("r")

    c6.setPen("c")

    c7.setPen(None)
    c7.setSymbol("d")
    c7.setSymbolSize(9)
    c7.setSymbolBrush("m")

    # move c5 and c7 to to y2
    w1vb2.addItem(c5)
    w1vb2.addItem(c7)

    sets = w1.getTrendSets()
    assert len(sets) == 4
    assert sets == [ts0, ts1, ts2, ts3]
    assert len(w1) == 6
    assert w1[:] == [c2, c3, c4, c5, c6, c7]

    assert Counter(w1._cprop_tool.getModifiableItems().values()) == Counter(
        [c0, c1, c2, c3, c4, c5, c6, c7]
    )
    assert Counter(w1vb1.addedItems) == Counter(
        [ts0, ts1, ts2, ts3, c0, c2, c4, c6]
    )
    assert Counter(w1vb2.addedItems) == Counter([c1, c3, c5, c7])

    # Save a config file with everything in w1
    f2 = tmp_path / "trend2.pck"
    with open(str(f2), "wb") as ofile:
        w1.saveConfigFile(ofile=ofile)
    assert f2.exists()
    assert len(list(tmp_path.iterdir())) == 2

    # test loadConfigFile

    w2 = tpg.TaurusTrend()
    qtbot.addWidget(w2)

    w2vb1 = w2.getViewBox()
    w2vb2 = w2._y2

    # load the second config file (containing everything)

    with open(str(f2), "rb") as ifile:
        w2.loadConfigFile(ifile=ifile)

    w1_props = w1._cprop_tool._getCurveAppearanceProperties()
    w2_props = w2._cprop_tool._getCurveAppearanceProperties()

    assert len(w1_props) == 8
    assert len(w2_props) == 6
    for k, p_aft in w2_props.items():
        assert k in w1_props
        p_ini = w1_props[k]
        assert not p_ini.conflictsWith(p_aft, strict=True)

    # show_and_wait(qtbot, w1, w2)  # uncomment for visually checking

    # load the first config file (does not contain TS2 and TS3)
    with open(str(f1), "rb") as ifile:
        w2.loadConfigFile(ifile=ifile)

    # check that there are not leftovers from previous config
    sets = w2.getTrendSets()
    assert len(w2) == 2  # only the curves from ts0 and ts1
    assert len(sets) == 2  # only ts0 and ts1
    assert [ts.base_name() for ts in sets] == ["TS0", "TS1"]
    assert [ts.name() for ts in w2] == ["TS0[0]", "TS1[0]"]

    # note that c3 is correctly in Y2 (because the config properly restores it)
    assert Counter(w2vb1.addedItems) == Counter(sets + [w2[0]])
    assert Counter(w2vb2.addedItems) == Counter([w2[1]])

    w2_props = w2._cprop_tool._getCurveAppearanceProperties()
    assert len(w2_props) == 2  # only the curves from ts0 and ts1 !
    for k, p_aft in w2_props.items():
        assert k in w1_props
        p_ini = w1_props[k]
        assert not p_ini.conflictsWith(p_aft, strict=True)

    # show_and_wait(qtbot, w1, w2)  # uncomment for visual checks

    # avoid teardown issues
    w1.setModel(None)
    w2.setModel(None)


def test_multiple_setModel(qtbot):
    """
    Check that repeated calls to setModel do not duplicate the items
    in the plot
    """
    w = tpg.TaurusTrend()
    qtbot.addWidget(w)
    for i in range(5):
        w.setModel(["eval:rand(2)"])
        sets = w.getTrendSets()
        assert Counter(w.getPlotItem().listDataItems()) == Counter(
            [sets[0], w[0], w[1]]
        ), "Found duplicates after {} calls to setModel".format(i + 1)
    # workaround for teardown issue
    w.setModel(None)


# def test_autopan(qtbot):
#     import taurus
#
#     taurus.changeDefaultPollingPeriod(222)
#     w = tpg.TaurusTrend()
#     qtbot.addWidget(w)
#
#     w.setModel(["eval:rand()"])
#     tpg.set_y_axis_for_curve(True, w[0], w.getPlotItem(), w._y2)
#     w._autopan.toggle()
#
#     show_and_wait(qtbot, w)
#
#     w.setModel(None)


def test_duplicate_curve_titles(qtbot):
    """
    Check that curves with same title are changed to have unique titles
    """
    w = tpg.TaurusTrend()
    qtbot.addWidget(w)
    w.setModel(
        [
            "eval:@foo/a=0+rand(2);a",
            "eval:@bar/a=1+rand(2);a",
            (None, "eval:2+rand()", "a"),
            "eval:3+rand(2)",
        ]
    )

    assert len(w[:]) == 7
    mod_items = w._cprop_tool.getModifiableItems()
    assert len(mod_items) == 7, "{}".format(mod_items)
    assert list(sorted(mod_items.keys())) == [
        "3+rand(2)[0]",
        "3+rand(2)[1]",
        "a[0]",
        "a[0] (2)",
        "a[0] (3)",
        "a[1]",
        "a[1] (2)",
    ]
    # show_and_wait(qtbot, w)
    w.setModel(None)



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################
__all__ = ["BufferSizeTool"]

from taurus.external.qt import QtGui, QtCore
from taurus.qt.qtcore.configuration.configuration import BaseConfigurableClass


class BufferSizeTool(QtGui.QAction, BaseConfigurableClass):
    """
    This tool provides a menu option to control the "Maximum buffer" of
    Plot data items that implement a `setBufferSize` method
    (see, e.g. :meth:`TaurusTrendSet.setBufferSize`).

    This tool inserts an action with a spinbox and emits a `valueChanged`
    signal whenever the value is changed.

    The connection between the data items and this tool can be done manually
    (by connecting to the `valueChanged` signal or automatically, if
    :meth:`autoconnect()` is `True` (default). The autoconnection feature works
    by discovering the compliant data items that are associated to the
    plot_item.

    This tool is implemented as an Action, and provides a method to attach it
    to a :class:`pyqtgraph.PlotItem`
    """

    valueChanged = QtCore.pyqtSignal(int)

    def __init__(
        self,
        parent=None,
        buffer_size=65536,
        text="Change buffers size...",
        autoconnect=True,
    ):
        BaseConfigurableClass.__init__(self)
        QtGui.QAction.__init__(self, text, parent)
        self.setToolTip("Maximum number of points for each trend")
        self._maxSize = buffer_size
        self._autoconnect = autoconnect

        # register config properties
        self.registerConfigProperty(
            self.bufferSize, self.setBufferSize, "buffer_size"
        )
        self.registerConfigProperty(
            self.autoconnect, self.setAutoconnect, "autoconnect"
        )

        # internal conections
        self.triggered.connect(self._onTriggered)

    def _onEdittingFinished(self):
        """
        emit valueChanged and update all associated trendsets (if
        self.autoconnect=True
        """
        buffer_size = self.bufferSize()
        self.valueChanged.emit(buffer_size)
        if self.autoconnect() and self.plot_item is not None:
            for item in self.plot_item.listDataItems():
                if hasattr(item, "setBufferSize"):
                    item.setBufferSize(buffer_size)

    def _onTriggered(self):
        maxSize = self.bufferSize()
        maxSize, ok = QtGui.QInputDialog.getInt(
            self.parentWidget(),
            "New buffer data size",
            "Enter the number of points to be kept in memory for each curve",
            maxSize,
            2,
            10000000,
            1000,
        )
        if ok:
            self.setBufferSize(maxSize)

    def attachToPlotItem(self, plot_item):
        """Use this method to add this tool to a plot

        :param plot_item: (PlotItem)
        """
        menu = plot_item.getViewBox().menu
        menu.addAction(self)
        self.plot_item = plot_item
        # ensure that current items are set to max size
        self.setBufferSize(self.bufferSize())
        if self.autoconnect():
            # enable the buffer limit also for trendsets added in the future
            try:  # requires https://github.com/pyqtgraph/pyqtgraph/pull/1388
                plot_item.scene().sigItemAdded.connect(self._onAddedItem)
            except AttributeError:
                pass

    def _onAddedItem(self, item):
        if hasattr(item, "setBufferSize"):
            item.setBufferSize(self.bufferSize())

    def autoconnect(self):
        """Returns autoconnect state

        :return: (bool)
        """
        return self._autoconnect

    def setAutoconnect(self, autoconnect):
        """Set autoconnect state. If True, the tool will autodetect trendsets
        associated to the plot item and will call setBufferSize
        on each of them for each change. If False, it will only emit a
        valueChanged signal and only those connected to it will be notified
        of changes

        :param autoconnect: (bool)
        """
        self._autoconnect = autoconnect

    def bufferSize(self):
        """Returns the current buffer_size value

        :return: (int)
        """
        return self._maxSize

    def setBufferSize(self, buffer_size):
        """Change the buffer_size value.

        :param buffer_size: (int) buffer_size
        """
        self._maxSize = buffer_size
        # update existing items
        if self.autoconnect() and self.plot_item is not None:
            for item in self.plot_item.listDataItems():
                if hasattr(item, "setBufferSize"):
                    item.setBufferSize(buffer_size)
        # emit a valueChanged signal
        self.valueChanged.emit(buffer_size)


if __name__ == "__main__":
    import taurus

    # taurus.setLogLevel(taurus.Debug)
    taurus.changeDefaultPollingPeriod(333)

    import sys
    from taurus.qt.qtgui.application import TaurusApplication
    from taurus_pyqtgraph import TaurusTrendSet, DateAxisItem
    import pyqtgraph as pg

    # from taurus_pyqtgraph import MaxBufferTool

    app = TaurusApplication()

    w = pg.PlotWidget()

    axis = DateAxisItem(orientation="bottom")
    w = pg.PlotWidget()
    axis.attachToPlotItem(w.getPlotItem())

    # test adding the curve before the tool
    ts1 = TaurusTrendSet(name="before", symbol="o")
    ts1.setModel("eval:rand()+1")

    w.addItem(ts1)

    fr = BufferSizeTool(w, buffer_size=16)
    fr.attachToPlotItem(w.getPlotItem())

    # test adding the curve after the tool
    ts2 = TaurusTrendSet(name="after", symbol="+")
    ts2.setModel("eval:rand()")

    w.addItem(ts2)

    w.show()

    ret = app.exec_()

    import pprint

    pprint.pprint(fr.createConfig())

    sys.exit(ret)



from taurus.external.qt import QtWidgets

EVALUATION_KEYS = {
    "{attr.label}": lambda proxy: proxy.label,
    "{attr.name}": lambda proxy: proxy.name,
    "{attr.fullname}": lambda proxy: proxy.fullname,
    "{dev.name}": lambda proxy: proxy.parentObj.name,
    "{dev.fullname}": lambda proxy: proxy.parentObj.fullname
}


class TitlePatternEditor:

    def __init__(self):
        self.legend_pattern = "{attr.label}"

    def showDialog(self):
        """Show pop up to change title on curves"""
        msg = (
            "Choose the title format. \n"
            + "You may use Python format() syntax. The TaurusDevice object\n"
            + 'can be referenced as "dev" and the TaurusAttribute object\n'
            + 'as "attr"'
        )

        evaluation_keys = {
            "{attr.label}": lambda proxy: proxy.label,
            "{attr.name}": lambda proxy: proxy.name,
            "{attr.fullname}": lambda proxy: proxy.fullname,
            "{dev.name}": lambda proxy: proxy.parentObj.name,
            "{dev.fullname}": lambda proxy: proxy.parentObj.fullname
        }

        if self.legend_pattern not in evaluation_keys:
            evaluation_keys[self.legend_pattern] = lambda proxy: ""

        selected_option_index = (list(evaluation_keys.keys())
                                 .index(self.legend_pattern))

        label_config, ok = QtWidgets.QInputDialog.getItem(
            None, "Change Title Pattern", msg, evaluation_keys,
            selected_option_index, True
        )

        if ok:
            self.legend_pattern = label_config
            return label_config
        else:
            return False



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""Top-level package for taurus_pyqtgraph."""

from __future__ import absolute_import
import taurus.external.qt as _  # avoid API1 errors due to pyqtgraph imports
from .y2axis import Y2ViewBox, set_y_axis_for_curve
from .curvespropertiestool import CurvesPropertiesTool
from .dateaxisitem import DateAxisItem
from .autopantool import XAutoPanTool
from .plot import TaurusPlot
from .trend import TaurusTrend
from .legendtool import PlotLegendTool
from .forcedreadtool import ForcedReadTool
from .buffersizetool import BufferSizeTool
from .taurusimageitem import TaurusImageItem
from .taurusplotdataitem import TaurusPlotDataItem
from .taurustrendset import TaurusTrendSet, TrendCurve
from .curvesmodel import TaurusItemConf, TaurusItemConfDlg
from .taurusmodelchoosertool import (
    TaurusModelChooserTool,
    TaurusImgModelChooserTool,
    TaurusXYModelChooserTool,
)
from .curveproperties import (
    CurveAppearanceProperties,
    CurvesAppearanceChooser,
    serialize_opts,
    deserialize_opts,
)
from .datainspectortool import DataInspectorLine, DataInspectorTool
from .util import unique_data_item_name, ensure_unique_curve_name
from .exporters import Taurus4TextExporter

# Do not modify the __version__ manually. To be modified by bumpversion
__version__ = "0.9.3-alpha"



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################
__all__ = ["TaurusTrendSet", "TrendCurve"]

"""This provides the pyqtgraph implementation of :class:`TaurusTrendSet`"""

import copy
import numpy
from taurus.core import (TaurusEventType, TaurusTimeVal,
                         TaurusAttrValue, Quantity)
from taurus.qt.qtgui.base import TaurusBaseComponent
from taurus.core.util.containers import ArrayBuffer, LoopList
from taurus.external.qt import Qt
from pyqtgraph import PlotDataItem

from .forcedreadtool import ForcedReadTool
from .curveproperties import CURVE_COLORS
from .util import ensure_unique_curve_name

import taurus


class TrendCurve(PlotDataItem):
    """A PlotDataItem for representing 1D trends"""
    def __repr__(self):
        return "<TrendCurve {}>".format(self.name())


class TaurusTrendSet(PlotDataItem, TaurusBaseComponent):
    """
    A PlotDataItem for displaying trend curve(s) associated to a
    TaurusAttribute. The TaurusTrendSet itself does not contain any data,
    but acts as a manager that dynamically adds/removes curve(s) (other
    PlotDataItems) to its associated plot.

    If the attribute is a scalar, the Trend Set generates only one curve
    representing the evolution of the value of the attribute. If the attribute
    is an array, as many curves as the attribute size are created,
    each representing the evolution of the value of a component of the array.

    When an event is received, all curves belonging to a TaurusTrendSet
    are updated.

    TaurusTrendSet can be considered as a container of (sorted) curves.
    As such, the curves contained by it can be accessed by index::

        ts = TaurusTrendSet('eval:rand(3)')
        # (...) wait for a Taurus Event arriving so that the curves are created
        ncurves = len(ts)  # ncurves will be 3 (assuming the event arrived)
        curve0 = ts[0]     # you can access the curve by index


    Note that internally each curve is a :class:`pyqtgraph.PlotDataItem` (i.e.,
    it is not aware of events by itself, but it relies on the TaurusTrendSet
    object to update its values)
    """

    def __init__(self, *args, **kwargs):
        _ = kwargs.pop("xModel", None)
        yModel = kwargs.pop("yModel", None)
        colors = kwargs.pop("colors", None)
        if colors is None:
            colors = LoopList(CURVE_COLORS)
        name = kwargs.pop("name", None)
        if isinstance(name, list):
            base_name = name[0]
            cnames = name[1:]
            if yModel is None:
                cnames = []
                self.warning("curve names ignored because yModel is not given")
        else:
            base_name, cnames = name, []
        PlotDataItem.__init__(self, x=[], y=[], name=base_name)
        TaurusBaseComponent.__init__(self, "TaurusBaseComponent")
        self._UImodifiable = False
        self._maxBufferSize = 65536  # (=2**16, i.e., 64K events))
        self._xBuffer = None
        self._yBuffer = None
        self._curveColors = colors
        self._args = args
        self._kwargs = kwargs
        self._curves = []
        self._timer = Qt.QTimer()
        self._timer.timeout.connect(self._forceRead)
        self._legend = None

        # register config properties
        self.setModelInConfig(True)
        self.registerConfigProperty(
            self._getCurveNames(), self._setCurveNames, "CurveNames"
        )
        self.registerConfigProperty(
            self._getCurvesOpts, self._setCurvesOpts, "opts"
        )
        # TODO: store forceReadPeriod config
        # TODO: store _maxBufferSize config
        if yModel is not None:
            self.setModel(yModel)
            if cnames:
                self._setCurveNames(cnames)

    def __repr__(self):
        return "<TrendSet {} ({} items)>".format(self.base_name(), len(self))

    def name(self):
        """Reimplemented from PlotDataItem to avoid having the ts itself added
        to legends.

        .. seealso:: :meth:`basename`
        """
        return None

    def base_name(self):
        """Returns the name of the trendset, which is used as a prefix for
        constructing the associated curves names

        .. seealso:: :meth:`name`
        """
        return PlotDataItem.name(self)

    def __getitem__(self, k):
        return self._curves[k]

    def __len__(self):
        return len(self._curves)

    def __contains__(self, k):
        return k in self._curves

    def setModel(self, name):
        """Reimplemented from :meth:`TaurusBaseComponent.setModel`"""
        TaurusBaseComponent.setModel(self, name)
        # force a read to ensure that the curves are created
        self._forceRead()

    def _initBuffers(self, ntrends):
        """initializes new x and y buffers"""

        self._yBuffer = ArrayBuffer(
            numpy.zeros((min(128, self._maxBufferSize), ntrends), dtype="d"),
            maxSize=self._maxBufferSize,
        )

        self._xBuffer = ArrayBuffer(
            (numpy.zeros(min(128, self._maxBufferSize), dtype="d")),
            maxSize=self._maxBufferSize,
        )

    def _initCurves(self, ntrends):
        """ Initializes new curves """

        # self._removeFromLegend(self._legend)

        # remove previously existing curves from views
        self._updateViewBox(None)

        self._curves = []

        if self._curveColors is None:
            self._curveColors = LoopList(CURVE_COLORS)
            self._curveColors.setCurrentIndex(-1)

        a = self._args
        kw = self._kwargs.copy()

        base_name = (
            self.base_name()
            or taurus.Attribute(self.getModel()).getSimpleName()
        )

        for i in range(ntrends):
            subname = "%s[%i]" % (base_name, i)
            kw["name"] = subname
            curve = TrendCurve(*a, **kw)
            if "pen" not in kw:
                curve.setPen(next(self._curveColors))
            self._curves.append(curve)
        self._updateViewBox(self.getViewBox())

    def _addToLegend(self, legend):
        # ------------------------------------------------------------------
        # In theory, TaurusTrendSet only uses viewBox.addItem to add its
        # sub-curves to the plot. In theory this should not add the curves
        # to the legend, and therefore we should do it here.
        # But somewhere the curves are already being added to the legend, and
        # if we re-add them here we get duplicated legend entries
        # TODO: Find where are the curves being added to the legend
        pass
        # if legend is None:
        #    return
        # for c in self._curves:
        #    legend.addItem(c, c.name())
        # -------------------------------------------------------------------

    def _removeFromLegend(self, legend):
        if legend is None:
            return
        for c in self._curves:
            legend.removeItem(c.name())

    def _updateViewBox(self, viewBox):
        """Add/remove the "extra" curves from the viewbox if needed"""
        for curve in self._curves:
            curve_viewBox = curve.getViewBox()
            if curve_viewBox is not None:
                plotItem = None
                viewWidget = curve_viewBox.getViewWidget()
                if viewWidget is not None:
                    plotItem = viewWidget.getPlotItem()
                curve_viewBox.removeItem(curve)
                if plotItem is not None:
                    plotItem.removeItem(curve)
            if viewBox is not None:
                plotItem = None
                viewWidget = viewBox.getViewWidget()
                if viewWidget is not None:
                    plotItem = viewWidget.getPlotItem()
                if plotItem is not None:
                    curve = ensure_unique_curve_name(curve, plotItem)
                    _cname = curve.name()
                    params = {"all trends": _cname}
                    plotItem.addItem(curve, params=params)

    def _updateBuffers(self, evt_value):
        """Update the x and y buffers with the new data. If the new data is
        not compatible with the existing buffers, the buffers are reset
        """
        rvalue = evt_value.rvalue
        try:
            # We use .magnitude here as a workaround for older pint affected by
            # https://github.com/hgrecco/pint/issues/509
            ntrends = numpy.size(rvalue.magnitude)
        except AttributeError:
            # support boolean values
            _v = numpy.array(rvalue)
            if _v.dtype is numpy.dtype("bool"):
                from taurus.core.units import Quantity
                rvalue = Quantity(_v)
                ntrends = _v.size
            else:
                raise

        if not self._isDataCompatible(rvalue, ntrends):
            self._initBuffers(ntrends)
            self._yUnits = rvalue.units
            self._initCurves(ntrends)

        try:
            self._yBuffer.append(rvalue.to(self._yUnits).magnitude)
        except Exception as e:
            self.warning(
                "Problem updating buffer Y (%s):%s", rvalue, e
            )
            evt_value = None

        try:
            self._xBuffer.append(evt_value.time.totime())
        except Exception as e:
            self.warning("Problem updating buffer X (%s):%s", evt_value, e)

        return self._xBuffer.contents(), self._yBuffer.contents()

    def _isDataCompatible(self, rvalue, ntrends):
        """
        Check that the new rvalue is compatible with the current data in the
        buffers. Check shape and unit compatibility.
        """
        if self._xBuffer is None or self._yBuffer is None:
            return False

        if rvalue.dimensionality != self._yUnits.dimensionality:
            return False

        current_trends = numpy.prod(self._yBuffer.contents().shape[1:])

        if ntrends != current_trends:
            return False

        return True

    def _addData(self, x, y):
        for i, curve in enumerate(self._curves):
            curve.setData(x=x, y=y[:, i])

    def clearBuffer(self):
        """Reset the buffered data"""
        self._initBuffers(len(self._curves))

    def handleEvent(self, evt_src, evt_type, evt_value):
        """Reimplementation of :meth:`TaurusBaseComponent.handleEvent`"""

        # model = evt_src if evt_src is not None else self.getModelObj()

        if (
            evt_value is None
            or not hasattr(evt_value, "rvalue")
            or evt_value.rvalue is None
        ):
            self.info("Invalid value. Ignoring.")
            return
        else:
            try:
                xValues, yValues = self._updateBuffers(evt_value)
            except Exception:
                # TODO: handle dropped events see: TaurusTrend._onDroppedEvent
                raise

        self._addData(xValues, yValues)

    def parentChanged(self):
        """Reimplementation of :meth:`PlotDataItem.parentChanged` to handle
        the change of the containing viewbox
        """
        PlotDataItem.parentChanged(self)

        self._updateViewBox(self.getViewBox())

        # update legend if needed
        try:
            legend = self.getViewWidget().getPlotItem().legend
        except Exception:
            legend = None
        if legend is not self._legend:
            self._removeFromLegend(self._legend)
            self._addToLegend(legend)
            self._legend = legend

        # Set period from ForcedReadTool (if found)
        try:
            for a in self.getViewBox().menu.actions():
                if isinstance(a, ForcedReadTool) and a.autoconnect():
                    self.setForcedReadPeriod(a.period())
                    break
        except Exception as e:
            self.debug("cannot set period from ForcedReadTool: %r", e)

    @property
    def forcedReadPeriod(self):
        """Returns the forced reading period (in ms). A value <= 0 indicates
        that the forced reading is disabled
        """
        return self._timer.interval()

    def setForcedReadPeriod(self, period):
        """
        Forces periodic reading of the subscribed attribute in order to show
        new points even if no events are received.
        It will create fake events as needed with the read value.
        It will also block the plotting of regular events when period > 0.

        :param period: (int) period in milliseconds. Use period<=0 to stop the
                       forced periodic reading
        """

        # stop the timer and remove the __ONLY_OWN_EVENTS filter
        self._timer.stop()
        filters = self.getEventFilters()
        if self.__ONLY_OWN_EVENTS in filters:
            filters.remove(self.__ONLY_OWN_EVENTS)
            self.setEventFilters(filters)

        # if period is positive, set the filter and start
        if period > 0:
            self.insertEventFilter(self.__ONLY_OWN_EVENTS)
            self._timer.start(period)

    def _forceRead(self, cache=False):
        """Forces a read of the associated attribute.

        :param cache: (bool) If True, the reading will be done with cache=True
                      but the timestamp of the resulting event will be replaced
                      by the current time. If False, no cache will be used at
                      all.
        """
        value = self.getModelValueObj(cache=cache)
        if value is None:
            # This is intended to show no data when no points are available.
            # We can get archiving values if available and needed.
            value = TaurusAttrValue()
            value.rvalue = Quantity(numpy.nan)
            value.time = TaurusTimeVal.now()
            self.warning("The attribute has no an actual value, check either "
                         "if the attribute exists or the device is down. "
                         "Or load data from archiving if needed.")
        if cache and value is not None:
            value = copy.copy(value)
            value.time = TaurusTimeVal.now()
        self.fireEvent(self, TaurusEventType.Periodic, value)

    def __ONLY_OWN_EVENTS(self, s, t, v):
        """An event filter that rejects all events except those that originate
        from this object
        """
        if s is self:
            return s, t, v
        else:
            return None

    def _getCurveNames(self):
        """returns the list of curve names (may have been changed by user!)"""
        return [c.name() for c in self._curves]

    def _setCurveNames(self, names):
        """restore options to curves"""
        # If no curves are yet created, force a read to create them
        if not self._curves:
            self._forceRead(cache=True)
        if len(names) != len(self._curves):
            self.warning("Cannot restore trend curve names (%s)", names)
        for curve, name in zip(self._curves, names):
            curve.setData(name=name)

    def _getCurvesOpts(self):
        """returns a list of serialized opts (one for each curve)"""
        from taurus_pyqtgraph import serialize_opts

        return [serialize_opts(copy.copy(c.opts)) for c in self._curves]

    def _setCurvesOpts(self, all_opts):
        """restore options to curves"""
        # If no curves are yet created, force a read to create them
        if not self._curves:
            self._forceRead(cache=True)
        # Check consistency in the number of curves
        if len(self._curves) != len(all_opts):
            self.warning(
                "Cannot apply curve options (mismatch in curves number)"
            )
            return
        from taurus_pyqtgraph import deserialize_opts

        for c, opts in zip(self._curves, all_opts):
            c.opts = deserialize_opts(opts)

            # This is a workaround for the following pyqtgraph's bug:
            # https://github.com/pyqtgraph/pyqtgraph/issues/531
            if opts["connect"] == "all":
                c.opts["connect"] = "all"
            elif opts["connect"] == "pairs":
                c.opts["connect"] = "pairs"
            elif opts["connect"] == "finite":
                c.opts["connect"] = "finite"

    def getFullModelNames(self):
        """
        Return a tuple of (None, fullmodelname) for API compatibility with
        :class:`TaurusPlotDataItem`.
        """
        return (None, self.getFullModelName())

    def setBufferSize(self, buffer_size):
        """sets the maximum number of points to store per trend curve

        :param buffer_size: (int) max number of points to store per trend curve
        """
        self._maxBufferSize = buffer_size
        try:
            if self._xBuffer is not None:
                # discard oldest data if needed for downsizing
                excess = len(self._xBuffer) - buffer_size
                if excess > 0:
                    self._xBuffer.moveLeft(excess)
                    self._xBuffer.resizeBuffer(buffer_size)
                # resize
                self._xBuffer.setMaxSize(buffer_size)
            if self._yBuffer is not None:
                # discard oldest data if needed for downsizing
                excess = len(self._yBuffer) - buffer_size
                if excess > 0:
                    self._yBuffer.moveLeft(excess)
                    self._yBuffer.resizeBuffer(buffer_size)
                # resize
                self._yBuffer.setMaxSize(buffer_size)
        except ValueError:
            self.info(
                "buffer downsizing  requested."
                + "Current contents will be discarded"
            )
            self.clearBuffer()

    def bufferSize(self):
        """returns the maximum number of points to be stored by the trends"""
        return self._maxBufferSize

    def _update(self):
        """Force repaint of the curves"""
        xValues = self._xBuffer.contents()
        yValues = self._yBuffer.contents()
        self._addData(xValues, yValues)


if __name__ == "__main__":
    import sys
    import pyqtgraph as pg
    from taurus.qt.qtgui.application import TaurusApplication
    from taurus_pyqtgraph import (
        # TaurusTrendSet,
        DateAxisItem,
        XAutoPanTool,
        # TaurusModelChooserTool,
        CurvesPropertiesTool,
    )

    from taurus.core.taurusmanager import TaurusManager

    taurusM = TaurusManager()
    taurusM.changeDefaultPollingPeriod(1000)  # ms

    app = TaurusApplication()

    # a standard pyqtgraph plot_item
    axis = DateAxisItem(orientation="bottom")
    w = pg.PlotWidget()
    axis.attachToPlotItem(w.getPlotItem())

    cp = CurvesPropertiesTool()
    cp.attachToPlotItem(w.getPlotItem())

    autopan = XAutoPanTool()
    autopan.attachToPlotItem(w.getPlotItem())

    # add legend to the plot, for that we have to give a name to plot items
    w.addLegend()

    # adding a taurus data item...
    c2 = TaurusTrendSet(name="foo")
    c2.setModel("eval:rand(2)")
    # c2.setForcedReadPeriod(500)

    w.addItem(c2)

    # ...and remove it after a while
    def rem():
        w.removeItem(c2)

    Qt.QTimer.singleShot(2000, rem)

    # modelchooser = TaurusModelChooserTool(itemClass=TaurusTrendSet)
    # modelchooser.attachToPlotItem(w.getPlotItem())

    w.show()

    ret = app.exec_()

    # import pprint

    # pprint.pprint(c2.createConfig())

    sys.exit(ret)



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

__all__ = ["StatisticsTool"]

from datetime import datetime

import numpy as np
from taurus.external.qt import Qt
from taurus.external.qt.QtWidgets import QCheckBox, QVBoxLayout, \
    QSplitter, QWidget, QGridLayout, QTableWidget, QTableWidgetItem, \
    QPushButton, QLabel, QDateTimeEdit
from taurus.qt.qtgui.container import TaurusWidget
from taurus_pyqtgraph.taurusplotdataitem import TaurusPlotDataItem


class StatisticsTool(Qt.QAction):
    """
    This tool inserts an action in the menu of the :class:`pyqtgraph.PlotItem`
    to which it is attached to show statistics like
    min, max, mean, std, rms of curves.It is implemented as an Action,
    and provides a method to attach it to a
    PlotItem.
    """

    def __init__(self, parent=None, itemClass=None):
        Qt.QAction.__init__(self, "Calculate statistics", parent)
        self.parent = parent
        self.triggered.connect(self._onTriggered)
        self.plotItem = None
        self.legend = None
        if itemClass is None:
            itemClass = TaurusPlotDataItem
        self.itemClass = itemClass

    def attachToPlotItem(self, plot_item, parentWidget=None):
        """
        Use this method to add this tool to a plot

        :param plot_item: (PlotItem)
        """
        self.plotItem = plot_item
        if self.plotItem.legend is not None:
            self.legend = self.plotItem.legend

        menu = self.plotItem.getViewBox().menu
        menu.addAction(self)
        self.setParent(parentWidget or menu)

    def _onTriggered(self):
        StatisticsToolDlg.display(self.parent, self.plotItem.listDataItems())


class StatisticsToolDlg(TaurusWidget):
    def __init__(self, parent, data_items):
        super().__init__()

        self.parent = parent
        self.dataItems = data_items

        self.layout = QVBoxLayout()

        self.splitter = QSplitter()

        self.setupXLimitsFragment()
        self.setupStatsPickerFragment()
        self.setupTableFragment()

        self.calculateButton = QPushButton('(re)calculate')
        self.calculateButton.clicked.connect(lambda _: self.fillTable())

        self.layout.addWidget(self.calculateButton)
        self.setLayout(self.layout)

    def setupXLimitsFragment(self):
        """Method to setup the x_limits fragment of the StatisticalToolDlg."""
        self.xLimitsWidget = QWidget()
        self.xLimitsLayout = QGridLayout()

        self.checkboxMin = QCheckBox('min')
        self.checkboxMin.clicked.connect(lambda _: self.inputMin.setDisabled(
            not self.checkboxMin.isChecked()))
        self.inputMin = QDateTimeEdit()
        self.inputMin.setDisplayFormat("yyyy-MM-dd'T'HH:mm:ss.zzz")
        self.inputMin.setDateTime(
            datetime.fromtimestamp(self._get_min_range()))
        self.inputMin.setDisabled(True)
        self.xLimitsLayout.addWidget(QLabel("X limits"))
        self.xLimitsLayout.addWidget(self.checkboxMin, 1, 0)
        self.xLimitsLayout.addWidget(self.inputMin, 1, 1)

        self.checkboxMax = QCheckBox('max')
        self.checkboxMax.clicked.connect(lambda _: self.inputMax.setDisabled(
            not self.checkboxMax.isChecked()))
        self.inputMax = QDateTimeEdit()
        self.inputMax.setDisplayFormat("yyyy-MM-dd'T'HH:mm:ss.zzz")
        self.inputMax.setDateTime(
            datetime.fromtimestamp(self._get_max_range()))
        self.inputMax.setDisabled(True)
        self.xLimitsLayout.addWidget(self.checkboxMax, 2, 0)
        self.xLimitsLayout.addWidget(self.inputMax, 2, 1)

        self.xLimitsWidget.setLayout(self.xLimitsLayout)

        self.layout.addWidget(self.xLimitsWidget)

    def setupStatsPickerFragment(self):
        """
        Method to setup the stats_picker fragment of the StatisticalToolDlg.
        """
        self.statsPickerWidget = QWidget()
        self.statsPickerLayout = QGridLayout()
        self.statsCheckboxes = []

        self.statsPickerLayout.addWidget(QLabel("Stats"), 0, 0)

        self.stats = ["points", "min", "max", "mean", "std", "rms"]
        for i, stat in enumerate(self.stats):
            check_box = QCheckBox(stat)
            check_box.setChecked(True)
            check_box.clicked.connect(
                lambda checked, col=i:
                self.tableWidget.setColumnHidden(col, not checked)
            )
            self.statsPickerLayout.addWidget(check_box, 1, i)
            self.statsCheckboxes.append(check_box)

        self.statsPickerWidget.setLayout(self.statsPickerLayout)
        self.layout.addWidget(self.statsPickerWidget)

    def setupTableFragment(self):
        """Method to setup the table fragment of the StatisticalToolDlg."""
        self.tableWidget = QTableWidget(0, 0)

        self.fillTable()

        self.layout.addWidget(self.tableWidget)

    def fillTable(self):
        """Method to fill the table of the StatisticalToolDlg."""
        start = float("-inf")
        end = float("inf")
        if self.checkboxMin.isChecked():
            start = float(
                datetime.fromisoformat(self.inputMin.text()).timestamp())
        else:
            self.inputMin.setDateTime(
                datetime.fromtimestamp(self._get_min_range()))

        if self.checkboxMax.isChecked():
            end = float(
                datetime.fromisoformat(self.inputMax.text()).timestamp())
        else:
            self.inputMax.setDateTime(
                datetime.fromtimestamp(self._get_max_range()))

        rows = [item for item in self.dataItems if item.name()]
        columns = []
        for idx, item in enumerate(self.statsCheckboxes):
            if item.isChecked():
                columns.append(self.stats[idx])

        self.tableWidget.setColumnCount(len(columns))
        self.tableWidget.setRowCount(len(rows))

        self.tableWidget.setHorizontalHeaderLabels(columns)
        self.tableWidget.setVerticalHeaderLabels(
            [item.name() for item in rows])

        for x in range(len(rows)):
            for y in range(len(columns)):
                item = rows[x]
                if item.xData is not None and item.yData is not None:
                    yData = [y for _, y in
                             filter(lambda item: start < item[0] < end,
                                    zip(item.xData, item.yData))]
                    xData = [x for x in
                             filter(lambda x: start < x < end, item.xData)]
                else:
                    yData, xData = None, None

                yStat = self.calcStat(columns[y], yData)
                if columns[y] in ["min", "max"]:
                    xStat = self.calcStat(columns[y], xData, time=True)
                    if xStat is not None:
                        self.tableWidget.setItem(x, y, QTableWidgetItem(
                            "t={}\ny={}".format(xStat, yStat)))
                    else:
                        self.tableWidget.setItem(x, y, QTableWidgetItem(""))
                else:
                    if yStat is not None:
                        self.tableWidget.setItem(x, y, QTableWidgetItem(
                            "{:g}".format(yStat)))
                    else:
                        self.tableWidget.setItem(x, y, QTableWidgetItem(""))
        self.tableWidget.resizeRowsToContents()

    def calcStat(self, stat, data, time=False):
        """Method to calculate the statistics for the given data."""
        if data is None or not len(data):
            return None
        else:
            return {
                'points': len(data),
                'min': datetime.fromtimestamp(
                    min(data)).strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3]
                if time else min(data),
                'max': datetime.fromtimestamp(
                    max(data)).strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3]
                if time else max(data),
                'mean': np.mean(data),
                'std': np.std(data) if len(data) >= 2 else None,
                'rms': np.sqrt(np.mean(np.array(data) ** 2)),
            }[stat]

    def _get_min_range(self):
        return min([min(curve.xData) for curve in self.dataItems if
                    curve.xData is not None])

    def _get_max_range(self):
        return max([max(curve.xData) for curve in self.dataItems if
                    curve.xData is not None])

    @staticmethod
    def display(parent, data_items):
        """Static method to display the StatisticalToolDlg."""
        dlg = Qt.QDialog(parent)
        dlg.setWindowTitle("Curve Stats Dialog")
        dlg.setWindowIcon(Qt.QIcon("logos:taurus.png"))
        layout = Qt.QVBoxLayout()
        w = StatisticsToolDlg(
            parent=parent,
            data_items=data_items
        )
        layout.addWidget(w)
        dlg.setLayout(layout)
        dlg.exec_()


def _demo_Statisticaltool():
    import sys
    import numpy
    import pyqtgraph as pg
    from taurus.qt.qtgui.application import TaurusApplication
    from taurus_pyqtgraph import TaurusPlotDataItem

    app = TaurusApplication()

    # a standard pyqtgraph plot_item
    w = pg.PlotWidget()

    # add legend to the plot, for that we have to give a name to plot items
    w.addLegend()

    # adding a regular data item (non-taurus)
    c1 = pg.PlotDataItem(name="st plot", pen="b", fillLevel=0, brush="c")
    c1.setData(numpy.arange(300) / 300.0)
    w.addItem(c1)

    # adding a taurus data item
    c2 = TaurusPlotDataItem(name="st2 plot", pen="r", symbol="o")
    c2.setModel("eval:rand(222)")

    w.addItem(c2)

    # attach to plot item
    tool = StatisticsTool(itemClass=TaurusPlotDataItem)
    tool.attachToPlotItem(w.getPlotItem())

    w.show()

    tool.trigger()

    sys.exit(app.exec_())


if __name__ == "__main__":
    _demo_Statisticaltool()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

__all__ = ["Y2ViewBox", "set_y_axis_for_curve"]


from pyqtgraph import ViewBox, PlotItem
from taurus.external.qt import QtWidgets

from taurus.qt.qtcore.configuration.configuration import BaseConfigurableClass


def set_y_axis_for_curve(y2, dataItem, plotItem, y2Axis):
    """
    Sets properties provided in the `properties` dict to curves provided in
    the `curves` dict. The association of a given curve with a property is
    done by matching the keys in the respective dictionaries
    :param y2: `True` indicates that the `dataItem` should be associated to y2,
               `False` indicates that it should be associated to the main
               viewbox, and `None` indicates that no change should be done.
    :param plotItem: The :class:`PlotItem` containing the dataItem.
    :param y2Axis: The :class:`Y2ViewBox` instance
    """
    # Set the Y1 / Y2 axis if required
    old_view = dataItem.getViewBox()  # current view for the curve
    if y2 is None:  # axis is not to be changed
        new_view = old_view
    elif y2:  # Y axis must be Y2
        new_view = y2Axis  # y2 axis view
    else:  # Y axis must be Y1
        new_view = plotItem.getViewBox()  # main view

    if new_view is not old_view:
        if old_view is not None:
            old_view.removeItem(dataItem)
        if not y2:
            # adapt the log mode to the main view logMode
            # (this is already done automatically when adding to y2)
            dataItem.setLogMode(
                plotItem.getAxis("bottom").logMode,
                plotItem.getAxis("left").logMode,
            )
        new_view.addItem(dataItem)
        old_view.autoRange()
        new_view.autoRange()


def _PlotItem_addItem(self, item, *args, **kwargs):
    """replacement for `PlotItem.addItem` that Y2Axis will use to monkey-patch
    the original one
    """
    PlotItem.addItem(self, item, *args, **kwargs)

    if hasattr(item, "setLogMode"):
        item.setLogMode(
            self.getAxis("bottom").logMode, self.getAxis("left").logMode
        )


class Y2ViewBox(ViewBox, BaseConfigurableClass):
    """
    A tool that inserts a secondary Y axis to a plot item (see
    :meth:`attachToPlotItem`).
    It is implemented as a :class:`pyqtgraph.ViewBox` and provides methods to
    add and remove :class:`pyqtgraph.PlotDataItem` objects to it.
    """

    def __init__(self, *args, **kwargs):
        self._isAttached = False
        self.plotItem = None
        name = kwargs.pop("name", "Y2 ViewBox")
        BaseConfigurableClass.__init__(self)
        ViewBox.__init__(self, *args, name=name, **kwargs)

        self.registerConfigProperty(
            self._getCurvesNames, self._addCurvesByName, "Y2Curves"
        )
        self.registerConfigProperty(self._getState, self.setState, "viewState")

    def attachToPlotItem(self, plot_item):
        """Use this method to add this axis to a plot

        :param plot_item: (PlotItem)
        """
        if self._isAttached:
            return  # TODO: log a message it's already attached
        self._isAttached = True

        mainViewBox = plot_item.getViewBox()
        mainViewBox.sigResized.connect(self._updateViews)

        self.plotItem = plot_item

        # add axis-independent actions for logarithmic scale
        self._addLogAxisActions()
        # disable the standard (custom view-unfriendly) log actions
        self.plotItem.ctrl.logXCheck.setEnabled(False)
        self.plotItem.ctrl.logYCheck.setEnabled(False)

        # monkey-patch the addItem method of the PlotItem
        from types import MethodType

        self.plotItem.addItem = MethodType(_PlotItem_addItem, self.plotItem)

        # add Y2 to main scene(), show the axis and link X axis to self.
        # self.plotItem.showAxis("right", show=bool(self.addedItems))
        self.plotItem.scene().addItem(self)
        self.plotItem.getAxis("right").linkToView(self)
        self.setXLink(self.plotItem.getViewBox())

        # make autorange button work for Y2 too
        self.plotItem.autoBtn.clicked.connect(self._onAutoBtnClicked)

    def _updateViews(self, viewBox):
        self.setGeometry(viewBox.sceneBoundingRect())
        self.linkedViewChanged(viewBox, self.XAxis)

    def removeItem(self, item):
        """Reimplemented from :class:`pyqtgraph.ViewBox`"""
        ViewBox.removeItem(self, item)
        if self.plotItem is not None:
            self.plotItem.showAxis("right", show=bool(self.addedItems))

    def addItem(self, item, ignoreBounds=False):
        """Reimplemented from :class:`pyqtgraph.ViewBox`"""

        # first add it to plotItem and then move it from main viewbox to y2
        if self.plotItem is not None:
            if item not in self.plotItem.listDataItems():
                self.plotItem.addItem(item)
            if item in self.plotItem.getViewBox().addedItems:
                self.plotItem.getViewBox().removeItem(item)
        ViewBox.addItem(self, item, ignoreBounds=ignoreBounds)

        if self.plotItem is not None:
            self.plotItem.showAxis("right", show=bool(self.addedItems))

            # set the item log mode to match this view:
            if hasattr(item, "setLogMode"):
                item.setLogMode(
                    self.plotItem.getAxis("bottom").logMode,
                    self.plotItem.getAxis("right").logMode,
                )

    def _getCurvesNames(self):
        """Returns the curve names associated to the Y2 axis.

        :return: (list) List of tuples of model names (xModelName, yModelName)
                 from each curve in this view
        """
        if self.plotItem is None:
            return []
        ret = []
        for c in self.plotItem.listDataItems():
            if c.getViewBox() == self and hasattr(c, "getFullModelNames"):
                ret.append(c.getFullModelNames())
        return ret

    def _addCurvesByName(self, names):
        curves = {}
        for c in self.plotItem.listDataItems():
            if hasattr(c, "getFullModelNames"):
                curves[c.getFullModelNames()] = c
        for n in names:
            c = curves[n]
            vb = c.getViewBox()
            if vb != self:
                vb.removeItem(c)
                self.addItem(c)

    def _getState(self):
        """Same as ViewBox.getState but removing viewRange conf to force
        a refresh with targetRange when loading
        """
        state = self.getState(copy=True)
        del state["viewRange"]
        return state

    def clearItems(self):
        """Remove the added items"""
        for c in self.addedItems:
            self.removeItem(c)

    def _addLogAxisActions(self):
        # insert & connect actions Log Scale Actions
        # In newer PyQtGraph versions (>0.13.1) there is no "menu.axes"
        # property, yet in older versions the QMenu is not added in
        # as child of the menu.

        def _addLogAxisAction(menu, position, on_toggle_slot):
            action = menu.addAction("Log scale")
            action.setCheckable(True)
            action.setChecked(self.plotItem.getAxis(position).logMode)
            action.setParent(menu)
            action.toggled.connect(on_toggle_slot)
            return action

        vb = self.plotItem.getViewBox()
        vb_submenus = vb.menu.findChildren(QtWidgets.QMenu)
        # if the list is empty, it is assumed the version is <0.13.2
        if not vb_submenus:
            # X (bottom)
            menu = self.plotItem.getViewBox().menu.axes[0]
            action = _addLogAxisAction(menu, "bottom", self._onXLogToggled)
            self.menu.axes[0].addAction(action)  # Add same action to X2 menu
            # Y1 (left)
            menu = self.plotItem.getViewBox().menu.axes[1]
            _addLogAxisAction(menu, "left", self._onY1LogToggled)
            # Y2 (right)
            menu = self.menu.axes[1]
            _addLogAxisAction(menu, "right", self._onY2LogToggled)
        else:
            for submenu in vb_submenus:
                if submenu.title() == "X axis":
                    action = _addLogAxisAction(submenu, "bottom",
                                               self._onXLogToggled)
                elif submenu.title() == "Y axis":
                    _addLogAxisAction(submenu, "left", self._onY1LogToggled)
            for submenu in self.menu.findChildren(QtWidgets.QMenu):
                if submenu.title() == "X axis":
                    submenu.addAction(action)  # Add same action to X2 menu too
                elif submenu.title() == "Y axis":
                    _addLogAxisAction(submenu, "right", self._onY2LogToggled)

    def _onXLogToggled(self, checked):
        logx = checked

        # set log mode for items of main viewbox
        logy = self.plotItem.getAxis("left").logMode
        for i in self.plotItem.getViewBox().addedItems:
            if hasattr(i, "setLogMode"):
                i.setLogMode(logx, logy)
        # set log mode for items of Y2 viewbox
        logy = self.plotItem.getAxis("right").logMode
        for i in self.addedItems:
            if hasattr(i, "setLogMode"):
                i.setLogMode(logx, logy)
        # set log mode for the bottom axis
        self.plotItem.getAxis("bottom").setLogMode(checked)

    def _onY1LogToggled(self, checked):
        # set log mode for items of main viewbox
        logx = self.plotItem.getAxis("bottom").logMode
        logy = checked
        for i in self.plotItem.getViewBox().addedItems:
            if hasattr(i, "setLogMode"):
                i.setLogMode(logx, logy)
        # set log mode for the left axis
        self.plotItem.getAxis("left").setLogMode(checked)

    def _onY2LogToggled(self, checked):
        # set log mode for items of Y2 viewbox
        logx = self.plotItem.getAxis("bottom").logMode
        logy = checked
        for i in self.addedItems:
            if hasattr(i, "setLogMode"):
                i.setLogMode(logx, logy)
        # set log mode for the right axis
        self.plotItem.getAxis("right").setLogMode(checked)

    def _onAutoBtnClicked(self):
        self.enableAutoRange()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

from __future__ import absolute_import

__all__ = ["TaurusPlot"]

import copy
import traceback
from taurus.external.qt import QtGui, PYSIDE2
from taurus.core.util.containers import LoopList
from taurus.core.util.log import Logger
from taurus.qt.qtcore.configuration import BaseConfigurableClass

from pyqtgraph import PlotWidget

from .curvespropertiestool import CurvesPropertiesTool
from .taurusmodelchoosertool import TaurusXYModelChooserTool
from .legendtool import PlotLegendTool
from .datainspectortool import DataInspectorTool
from .y2axis import Y2ViewBox
from .curveproperties import CURVE_COLORS
from .titlepatterneditor import TitlePatternEditor, EVALUATION_KEYS


class TaurusPlot(PlotWidget, BaseConfigurableClass):
    """
    TaurusPlot is a general widget for plotting 1D data sets. It is an extended
    taurus-aware version of :class:`pyqtgraph.PlotWidget`.

    Apart from all the features already available in a regulat PlotWidget,
    TaurusPlot incorporates the following tools/features:

        - Secondary Y axis (right axis)
        - A plot configuration dialog, and save/restore configuration
          facilities
        - A menu option for adding/removing models
        - A menu option for showing/hiding the legend
        - Automatic color change of curves for newly added models

    """

    def __init__(self, parent=None, **kwargs):
        if PYSIDE2:
            # Workaround for https://bugreports.qt.io/browse/PYSIDE-1564
            BaseConfigurableClass.__init__(self)
            PlotWidget.__init__(self, parent=parent, **kwargs)
        else:
            super(TaurusPlot, self).__init__(parent=None, **kwargs)

        # Compose with a Logger
        self._logger = Logger(name=self.__class__.__name__)
        self.debug = self._logger.debug
        self.info = self._logger.info
        self.warning = self._logger.warning
        self.error = self._logger.error

        # set up cyclic color generator
        self._curveColors = LoopList(CURVE_COLORS)
        self._curveColors.setCurrentIndex(-1)

        self.titlePatternEditor = TitlePatternEditor()

        # add save & retrieve configuration actions
        menu = self.getPlotItem().getViewBox().menu
        saveConfigAction = QtGui.QAction("Save configuration", menu)
        saveConfigAction.triggered.connect(self._onSaveConfigAction)
        menu.addAction(saveConfigAction)

        loadConfigAction = QtGui.QAction("Retrieve saved configuration", menu)
        loadConfigAction.triggered.connect(self._onRetrieveConfigAction)
        menu.addAction(loadConfigAction)

        # add change curve labels
        changeCurvesTitlesAction = QtGui.QAction("Change Curves Titles "
                                                 "(All curves)...", menu)
        changeCurvesTitlesAction.triggered.connect(
            self._onChangeCurvesTitlesAction)
        menu.addAction(changeCurvesTitlesAction)

        self.registerConfigProperty(self._getState, self.restoreState, "state")

        # add legend tool
        legend_tool = PlotLegendTool(self)
        legend_tool.attachToPlotItem(self.getPlotItem())

        # add model chooser
        self._model_chooser_tool = TaurusXYModelChooserTool(self)
        self._model_chooser_tool.attachToPlotItem(
            self.getPlotItem(), self, self._curveColors
        )

        # add Y2 axis
        self._y2 = Y2ViewBox()
        self._y2.attachToPlotItem(self.getPlotItem())

        # add plot configuration dialog
        self._cprop_tool = CurvesPropertiesTool(self)
        self._cprop_tool.attachToPlotItem(self.getPlotItem(), y2=self._y2)

        # add a data inspector
        inspector_tool = DataInspectorTool(self)
        inspector_tool.attachToPlotItem(self.getPlotItem())

        # enable Autorange
        self.getPlotItem().getViewBox().enableAutoRange(True)
        self._y2.enableAutoRange(True)

        # Register config properties
        self.registerConfigDelegate(self._model_chooser_tool, "XYmodelchooser")
        self.registerConfigDelegate(self._y2, "Y2Axis")
        self.registerConfigDelegate(self._cprop_tool, "CurvePropertiesTool")
        self.registerConfigDelegate(legend_tool, "legend")
        self.registerConfigDelegate(inspector_tool, "inspector")

    def __getitem__(self, idx):
        """
        Provides a list-like interface: items can be accessed using slice
        notation
        """
        return self.getPlotItem().listDataItems()[idx]

    def __len__(self):
        return len(self.getPlotItem().listDataItems())

    def __bool__(self):
        return True

    def setModel(self, names):
        """Reimplemented to delegate to the model chooser"""
        # support passing a string in names
        if isinstance(names, str):
            names = [names]
        self._model_chooser_tool.updateModels(names)

    def addModels(self, names):
        """Reimplemented to delegate to the  model chooser"""
        # support passing a string in names
        if isinstance(names, str):
            names = [names]
        self._model_chooser_tool.addModels(names)

    def _getState(self):
        """Same as PlotWidget.saveState but removing viewRange conf to force
        a refresh with targetRange when loading
        """
        state = copy.deepcopy(self.saveState())
        # remove viewRange conf
        del state["view"]["viewRange"]
        return state

    def setXAxisMode(self, x_axis_mode):
        """Required generic TaurusPlot API """
        from taurus_pyqtgraph import DateAxisItem

        if x_axis_mode == "t":
            axis = DateAxisItem(orientation="bottom")
            axis.attachToPlotItem(self.getPlotItem())
        elif x_axis_mode == "n":
            axis = self.getPlotItem().axes["bottom"]["item"]
            if isinstance(axis, DateAxisItem):
                axis.detachFromPlotItem()
        else:
            raise ValueError("Unsupported x axis mode {}".format(x_axis_mode))

    def _onSaveConfigAction(self):
        """wrapper to avoid issues with overloaded signals"""
        return self.saveConfigFile()

    def _onChangeCurvesTitlesAction(self):
        """Show pop up to change title on curves"""
        label_config = self.titlePatternEditor.showDialog()

        if label_config:
            for item in self.getPlotItem().curves:
                result = label_config
                for key, resolver in EVALUATION_KEYS.items():
                    if key in self.titlePatternEditor.legend_pattern:
                        try:
                            result = result.replace(key,
                                                    resolver(item.modelObj))
                        except AttributeError:
                            self._logger.warning(traceback.format_exc())
                            result = result.replace(key, "NoDef")

                if isinstance(item, list):
                    if len(item) == 1:
                        item[0].setData(name=result)
                        self.plotItem.legend.getLabel(item[0]).setText(result)

                    if len(item) > 1:
                        for i in range(len(item)):
                            result_array = "{}[{}]".format(result, i)
                            item[i].setData(name=result_array)
                            (self.plotItem.legend.getLabel(item[i])
                             .setText(result_array))
                else:
                    item.setData(name=result)
                    self.plotItem.legend.getLabel(item).setText(result)

    def _onRetrieveConfigAction(self):
        """wrapper to avoid issues with overloaded signals"""
        return self.loadConfigFile()


def plot_main(
    models=(),
    config_file=None,
    x_axis_mode="n",
    demo=False,
    window_name="TaurusPlot (pg)",
):
    """Launch a TaurusPlot"""
    import sys
    from taurus.qt.qtgui.application import TaurusApplication

    app = TaurusApplication(cmd_line_parser=None, app_name="taurusplot(pg)")

    w = TaurusPlot()

    # w.loadConfigFile('tmp/TaurusPlot.pck')

    w.setWindowTitle(window_name)

    if demo:
        models = list(models)
        models.extend(["eval:rand(100)", "eval:0.5*sqrt(arange(100))"])

    w.setXAxisMode(x_axis_mode.lower())

    if config_file is not None:
        w.loadConfigFile(config_file)

    if models:
        w.setModel(models)

    w.show()
    ret = app.exec_()

    # import pprint
    # pprint.pprint(w.createConfig())

    sys.exit(ret)


if __name__ == "__main__":
    plot_main()



import copy
from taurus.external.qt import Qt
from taurus_pyqtgraph.curveproperties import (
    CurveAppearanceProperties,
    CONFLICT,
)


p_all_conflict = CurveAppearanceProperties()


p1 = CurveAppearanceProperties(
    sStyle="t",
    sSize=2,
    sColor="r",
    sFill=False,
    lStyle=Qt.Qt.DashLine,
    lWidth=3,
    lColor="b",
    stepMode="left",
    y2=True,
    cFill=False,
    title="p1, all set",
    visible=True,
)

p2 = CurveAppearanceProperties(
    sStyle=None,
    sSize=-1,
    sColor="b",
    sFill=True,
    lStyle=Qt.Qt.SolidLine,
    lWidth=5,
    lColor="y",
    stepMode="right",
    y2=False,
    cFill=True,
    title="p2, all set",
    visible=False,
)

p3 = copy.deepcopy(p1)
p3.sSize = 3
p3.lStyle = CONFLICT

p4 = copy.deepcopy(p1)
p4.sSize = 3
p4.lColor = "g"


def test_curveproperties_conflictsWith(qtbot):
    assert p1.conflictsWith(copy.deepcopy(p1), strict=False) == []
    assert p1.conflictsWith(copy.deepcopy(p1), strict=True) == []
    assert p1.conflictsWith(p_all_conflict, strict=False) == []
    assert p1.conflictsWith(p_all_conflict, strict=True) == p1.propertyList
    assert p1.conflictsWith(p2, strict=False) == p1.propertyList
    assert p1.conflictsWith(p2, strict=True) == p1.propertyList
    assert p1.conflictsWith(p3, strict=False) == ["sSize"]
    assert p1.conflictsWith(p3, strict=True) == ["sSize", "lStyle"]


def test_curveproperties_merge(qtbot):
    merged = p1.merge([p1, p3])
    for a in merged.propertyList:
        if a in ["sSize", "lStyle"]:
            assert getattr(merged, a) == CONFLICT, "mismatch: %s" % a
        else:
            assert getattr(merged, a) == getattr(p1, a), "mismatch: %s" % a

    merged = p1.merge([p1, p3, p4])
    for a in merged.propertyList:
        if a in ["sSize", "lStyle", "lColor"]:
            assert getattr(merged, a) == CONFLICT, "mismatch: %s" % a
        else:
            assert getattr(merged, a) == getattr(p1, a), "mismatch: %s" % a

    merged = p1.merge([p3, p1, p3, p3, p3, p3])
    for a in merged.propertyList:
        if a in ["sSize", "lStyle"]:
            assert getattr(merged, a) == CONFLICT, "mismatch: %s" % a
        else:
            assert getattr(merged, a) == getattr(p1, a), "mismatch: %s" % a

    merged = p1.merge([p1, p3], conflict=p1.inConflict_update_a)
    assert merged.sSize == p3.sSize, "update a with b if they differ"
    assert merged.lStyle == p1.lStyle, "do not update when b is in conflict"
    for a in merged.propertyList:
        if a not in ["sSize", "lStyle"]:
            assert getattr(merged, a) == getattr(p3, a), "mismatch: %s" % a

    merged = p1.merge([p1, p2], conflict=p1.inConflict_update_a)
    for a in merged.propertyList:
        assert getattr(merged, a) == getattr(p2, a), "mismatch: %s" % a

    merged = p1.merge([p1, p_all_conflict], conflict=p1.inConflict_update_a)
    for a in merged.propertyList:
        assert getattr(merged, a) == getattr(p1, a), "mismatch: %s" % a

    merged = p1.merge([p1, p_all_conflict], conflict=p1.inConflict_update_a)
    for a in merged.propertyList:
        assert getattr(merged, a) == getattr(p1, a), "mismatch: %s" % a



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################
__all__ = ["TaurusPlotDataItem"]

import copy
from taurus import Attribute
from taurus.core import TaurusEventType
from taurus.qt.qtgui.base import TaurusBaseComponent
from pyqtgraph import PlotDataItem
from taurus.core.util.containers import LoopList
from .curveproperties import CURVE_COLORS


class TaurusPlotDataItem(PlotDataItem, TaurusBaseComponent):
    """A taurus-ified PlotDataItem"""

    def __init__(self, *args, **kwargs):
        """
        Accepts same args and kwargs as PlotDataItem, plus:
        :param xModel: (str) Taurus model name for abscissas values.
                       Default=None
        :param yModel: (str) Taurus model name for ordinate values.
                       Default=None
        :param colors: (generator) An infinite generator of QColor objects
        """
        xModel = kwargs.pop("xModel", None)
        yModel = kwargs.pop("yModel", None)
        colors = kwargs.pop("colors", LoopList(CURVE_COLORS))
        if "pen" not in kwargs:
            kwargs["pen"] = next(colors)
        PlotDataItem.__init__(self, *args, **kwargs)
        TaurusBaseComponent.__init__(self, "TaurusBaseComponent")
        self._x = None
        self._y = None
        self.xModel = None
        if xModel is not None:
            self.setXModel(xModel)
        if yModel is not None:
            self.setModel(yModel)

        self.registerConfigProperty(self.getOpts, self.setOpts, "opts")
        self.setModelInConfig(True)
        self.registerConfigProperty(
            self.getXModelName, self.setXModel, "XModel"
        )

    def setXModel(self, xModel):
        """Sets the taurus model for the x component"""
        if self.xModel is not None:
            self.xModel.removeListener(self)

        if not xModel:
            self.xModel = None
            return

        self.xModel = Attribute(xModel)
        self.xModel.addListener(self)

    def getXModelName(self):
        """Returns the full name for the x component"""
        if self.xModel is None:
            return None
        else:
            return self.xModel.getFullName()

    def handleEvent(self, evt_src, evt_type, evt_value):
        """Reimplemented from :class:`TaurusBaseComponent`"""
        if evt_type not in (TaurusEventType.Change, TaurusEventType.Periodic):
            return
        yModel = self.getModelObj()
        if yModel == evt_src:
            if yModel is not None:
                try:
                    # workaround to avoid UnitStrippedWarning (see #89)
                    # TODO: add proper unit support to TaurusPlot
                    self._y = evt_value.rvalue.magnitude
                except AttributeError:
                    self._y = evt_value.rvalue
            else:
                self._y = None
        if self.xModel == evt_src:
            if self.xModel is not None:
                try:
                    # workaround to avoid UnitStrippedWarning (see #89)
                    # TODO: add proper unit support to TaurusPlot
                    self._x = evt_value.rvalue.magnitude
                except AttributeError:
                    self._x = evt_value.rvalue
            else:
                self._x = None
        try:
            self.setData(x=self._x, y=self._y)
        except Exception as e:
            self.debug("Could not set data. Reason: %r", e)

    def getOpts(self):
        """
        Returns the options for this DataItem as a serialized, pickable
        dictionary
        """
        from taurus_pyqtgraph import serialize_opts

        return serialize_opts(copy.copy(self.opts))

    def setOpts(self, opts):
        # creates QPainters (QPen or QBrush) from a pickle loaded file
        # for adapt the serialized objects into PlotDataItem properties

        from taurus_pyqtgraph import deserialize_opts

        self.opts = deserialize_opts(opts)

        # This is a workaround for the following pyqtgraph's bug:
        # https://github.com/pyqtgraph/pyqtgraph/issues/531
        if opts["connect"] == "all":
            self.opts["connect"] = "all"
        elif opts["connect"] == "pairs":
            self.opts["connect"] = "pairs"
        elif opts["connect"] == "finite":
            self.opts["connect"] = "finite"

    def getFullModelNames(self):
        """
        Get the full model names for the x and y components as an
        `(xname,yname)` tuple
        """
        return (self.getXModelName(), self.getFullModelName())


if __name__ == "__main__":
    import sys
    import numpy
    import pyqtgraph as pg
    from taurus.qt.qtgui.application import TaurusApplication

    # from taurus_pyqtgraph import TaurusPlotDataItem
    from taurus.external.qt import Qt
    from taurus_pyqtgraph import TaurusModelChooserTool

    app = TaurusApplication()

    # a standard pyqtgraph plot_item
    w = pg.PlotWidget()

    # add legend to the plot, for that we have to give a name to plot items
    w.addLegend()

    # adding a regular data item (non-taurus)
    c1 = pg.PlotDataItem(name="st plot", pen="b", fillLevel=0, brush="c")
    c1.setData(numpy.arange(300) / 300.0)
    w.addItem(c1)

    pen = pg.mkPen(color="r", style=4)
    brush = pg.mkBrush(color="b")
    brush.setStyle(3)

    # adding a taurus data item
    # c2 = TaurusPlotDataItem(name='st2 plot', pen='r', symbol='o')
    c2 = TaurusPlotDataItem(pen=pen, name="foo")
    # c2 = TaurusPlotDataItem()

    # c2.loadConfigFile('tmp/conf.cfg')

    c2.setModel('eval:Quantity(rand(256),"m")')
    # c2.setModel('sys/tg_test/1/wave')
    # c2.setModel(None)
    # c2.setXModel(None)

    # c2.setXModel('eval:Quantity(rand(256),"m")')

    w.addItem(c2)

    # ...and remove it after a while
    def rem():
        w.removeItem(c2)

    Qt.QTimer.singleShot(2000, rem)

    modelchooser = TaurusModelChooserTool(itemClass=TaurusPlotDataItem)
    modelchooser.attachToPlotItem(w.getPlotItem())

    w.show()

    res = app.exec_()

    # config = c2.createConfig()
    # print config

    # c2.saveConfigFile('tmp/conf.cfg')

    sys.exit(res)



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
This module provides date-time aware axis
"""

__all__ = ["DateAxisItem"]

# -------------------------------------------------------------------------
# There is a conflict problem with PyQt versions. Pyqtgraph imports his own
# library of PyQt, and Taurus too. So we have to import Qt from own version
# first as a workaround for forcing our own (as a workaround)
from taurus.external.qt import Qt  # noqa

# -------------------------------------------------------------------------

import numpy
from pyqtgraph import AxisItem
from datetime import datetime, timedelta
from time import mktime
from taurus.external.qt import QtWidgets

_AXIS_INDEX = {"bottom": 0, "top": 0, "left": 1, "right": 1, "X": 0, "Y": 1}


class DateAxisItem(AxisItem):
    """
    A tool that provides a date-time aware axis. It is implemented as an
    AxisItem that interprets positions as unix timestamps (i.e. seconds
    since 1970).

    The labels and the tick positions are dynamically adjusted depending
    on the range.

    It provides a  :meth:`attachToPlotItem` method to add it to a given
    PlotItem
    """

    # TODO: Document this class and methods
    # Max width in pixels reserved for each label in axis
    _pxLabelWidth = 80

    def __init__(self, *args, **kwargs):
        AxisItem.__init__(self, *args, **kwargs)
        self._oldAxis = None
        self._minDTE = None
        self._maxDTE = None
        self._setAxisRange = None
        self._manualRadio = None

    def tickValues(self, minVal, maxVal, size):
        """
        Reimplemented from PlotItem to adjust to the range and to force
        the ticks at "round" positions in the context of time units instead of
        rounding in a decimal base
        """

        maxMajSteps = int(size // self._pxLabelWidth)

        dx = maxVal - minVal
        majticks = []

        try:
            dt1 = datetime.fromtimestamp(minVal)
            dt2 = datetime.fromtimestamp(maxVal)
        except Exception as e:
            from taurus import warning

            warning("Invalid range in DateTime axis: %s", e)
            return [(dx, [])]

        if dx > 63072001:  # 3600s*24*(365+366) = 2 years (count leap year)
            d = timedelta(days=366)
            for y in range(dt1.year + 1, dt2.year + 1):
                dt = datetime(year=y, month=1, day=1)
                majticks.append(mktime(dt.timetuple()))

        elif dx > 5270400:  # 3600s*24*61 = 61 days
            d = timedelta(days=31)
            dt = (
                dt1.replace(day=1, hour=0, minute=0, second=0, microsecond=0)
                + d
            )
            while dt < dt2:
                # make sure that we are on day 1 (even if always sum 31 days)
                dt = dt.replace(day=1)
                majticks.append(mktime(dt.timetuple()))
                dt += d

        elif dx > 172800:  # 3600s24*2 = 2 days
            d = timedelta(days=1)
            dt = dt1.replace(hour=0, minute=0, second=0, microsecond=0) + d
            while dt < dt2:
                majticks.append(mktime(dt.timetuple()))
                dt += d

        elif dx > 7200:  # 3600s*2 = 2hours
            d = timedelta(hours=1)
            dt = dt1.replace(minute=0, second=0, microsecond=0) + d
            while dt < dt2:
                majticks.append(mktime(dt.timetuple()))
                dt += d

        elif dx > 1200:  # 60s*20 = 20 minutes
            d = timedelta(minutes=10)
            dt = (
                dt1.replace(
                    minute=(dt1.minute // 10) * 10, second=0, microsecond=0
                )
                + d
            )
            while dt < dt2:
                majticks.append(mktime(dt.timetuple()))
                dt += d

        elif dx > 120:  # 60s*2 = 2 minutes
            d = timedelta(minutes=1)
            dt = dt1.replace(second=0, microsecond=0) + d
            while dt < dt2:
                majticks.append(mktime(dt.timetuple()))
                dt += d

        elif dx > 20:  # 20s
            d = timedelta(seconds=10)
            dt = dt1.replace(second=(dt1.second // 10) * 10, microsecond=0) + d
            while dt < dt2:
                majticks.append(mktime(dt.timetuple()))
                dt += d

        elif dx > 2:  # 2s
            d = timedelta(seconds=1)
            # majticks = list(range(int(minVal), int(maxVal)))
            majticks = list(
                range(int(numpy.ceil(minVal)), int(numpy.ceil(maxVal)))
            )

        else:  # <2s , use standard implementation from parent
            return AxisItem.tickValues(self, minVal, maxVal, size)

        # print("majticks >: ", majticks)

        L = len(majticks)
        if L > maxMajSteps:
            if maxMajSteps == 0:
                majticks = []
            else:
                majticks = majticks[:: int(numpy.ceil(float(L) / maxMajSteps))]

        # print("majticks <: ", majticks)
        # print "----------------------------"

        return [(d.total_seconds(), majticks)]

    def tickStrings(self, values, scale, spacing):
        """Reimplemented from PlotItem to adjust to the range"""
        ret = []
        if not values:
            return []
        # rng = max(values)-min(values)
        # print('values: ', values)
        # print('scale: ', scale)
        # print('spacing: ', spacing)

        if spacing >= 31622400:  # = timedelta(days=366).total_seconds
            fmt = "%Y"

        elif spacing >= 2678400:  # = timedelta(days=31).total_seconds
            fmt = "%Y %b"

        elif spacing >= 86400:  # = timedelta(days = 1).total_seconds
            fmt = "%b/%d"

        elif spacing >= 3600:  # = timedelta(hours=1).total_seconds
            fmt = "%b/%d-%Hh"

        elif spacing >= 600:  # = timedelta(minutes=10).total_seconds
            fmt = "%H:%M"

        elif spacing >= 60:  # = timedelta(minutes=1).total_seconds
            fmt = "%H:%M"

        elif spacing >= 10:  # 10 s
            fmt = "%H:%M:%S"

        elif spacing >= 1:  # 1s
            fmt = "%H:%M:%S"

        else:
            # less than 2s (show microseconds)
            # fmt = '%S.%f"'
            fmt = "[+%fms]"  # explicitly relative to last second

        for x in values:
            try:
                t = datetime.fromtimestamp(x)
                ret.append(t.strftime(fmt))
            except ValueError:  # Windows can't handle dates before 1970
                ret.append("")

        return ret

    def attachToPlotItem(self, plotItem):
        """Add this axis to the given PlotItem

        :param plotItem: (PlotItem)
        """
        self.setParentItem(plotItem)
        viewBox = plotItem.getViewBox()
        self.linkToView(viewBox)
        self._oldAxis = plotItem.axes[self.orientation]["item"]
        self._oldAxis.hide()
        plotItem.axes[self.orientation]["item"] = self
        pos = plotItem.axes[self.orientation]["pos"]
        plotItem.layout.addItem(self, *pos)
        self.setZValue(-1000)

        # initialize the datetime editors for the axis menu widgets
        self._minDTE = Qt.QDateTimeEdit()
        self._maxDTE = Qt.QDateTimeEdit()
        self._minDTE.setCalendarPopup(True)
        self._minDTE.setDisplayFormat("yyyy-MM-dd hh:mm:ss")
        self._maxDTE.setCalendarPopup(True)
        self._maxDTE.setDisplayFormat("yyyy-MM-dd hh:mm:ss")
        self._minDTE.dateTimeChanged.connect(self._maxDTE.setMinimumDateTime)
        self._maxDTE.dateTimeChanged.connect(self._minDTE.setMaximumDateTime)

        # replace the float-based range setting editors by datetime editors
        axis = _AXIS_INDEX[self.orientation]
        ctrl = viewBox.menu.ctrl[axis]
        ly = ctrl.gridLayout
        xmin_pos = ly.getItemPosition(ly.indexOf(ctrl.minText))
        xmax_pos = ly.getItemPosition(ly.indexOf(ctrl.maxText))
        ctrl.minText.hide()
        ctrl.maxText.hide()
        ly.addWidget(self._minDTE, *xmin_pos)
        ly.addWidget(self._maxDTE, *xmax_pos)

        # give room to the datetime editors in the axis menu widget;
        # In newer PyQtGraph versions (>0.13.1) there is no "menu.axes"
        # property, yet in older versions the QMenu is not added in
        # menu.children().
        # if the list is empty, it is assumed the version is <0.13.2
        for submenu in viewBox.menu.findChildren(QtWidgets.QMenu):
            axis_name = ('X axis' if axis == 0 else 'Y axis')
            if submenu.title() == axis_name:
                w = submenu.actions()[0].defaultWidget()
                break
        else:
            w = viewBox.menu.axes[axis].actions()[0].defaultWidget()
        w.setMaximumWidth(max(450, w.maximumWidth()))

        # Update the datetime editors and connect them to view changes
        self._updateMenu()
        viewBox.sigStateChanged.connect(self._updateMenu)

        # connect the datetime changed signals to _onRangeEdited
        self._setAxisRange = [viewBox.setXRange, viewBox.setYRange][axis]
        self._minDTE.dateTimeChanged.connect(self._onRangeEdited)
        self._maxDTE.dateTimeChanged.connect(self._onRangeEdited)

        # keep a reference to the manual radio button
        self._manualRadio = ctrl.manualRadio

    def detachFromPlotItem(self):
        """Remove this axis from its attached PlotItem
        (not yet implemented)
        """
        pass  # TODO

    def _onRangeEdited(self):
        """Change the range when the datetime editors qre edited"""
        # avoid infinite recursion by only updating when the menu is shown
        if not self._minDTE.isVisible():
            return
        self._manualRadio.setChecked(True)
        self._setAxisRange(
            self._minDTE.dateTime().toMSecsSinceEpoch() / 1000,
            self._maxDTE.dateTime().toMSecsSinceEpoch() / 1000,
            padding=0,
        )

    def _updateMenu(self):
        """Update the datetime editors when the range is changes"""
        # avoid infinite recursion (and confusion) by only updating when the
        # menu is not visible
        if self._minDTE.isVisible():
            return
        view = self.linkedView()
        state = view.getState(copy=False)
        axis = _AXIS_INDEX[self.orientation]
        _min, _max = state["targetRange"][axis]
        _min = int(numpy.floor(_min))
        _max = int(numpy.ceil(_max))
        self._minDTE.setDateTime(Qt.QDateTime.fromMSecsSinceEpoch(_min * 1000))
        self._maxDTE.setDateTime(Qt.QDateTime.fromMSecsSinceEpoch(_max * 1000))


if __name__ == "__main__":

    import sys
    import pyqtgraph as pg
    from taurus.qt.qtgui.application import TaurusApplication
    from taurus_pyqtgraph import TaurusPlotDataItem

    app = TaurusApplication()

    # a standard pyqtgraph plot_item
    w = pg.PlotWidget()
    axis = DateAxisItem(orientation="bottom")

    axis.attachToPlotItem(w.getPlotItem())

    # adding a taurus data item
    c2 = TaurusPlotDataItem()
    w.addItem(c2)

    w.show()

    sys.exit(app.exec_())



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################


"""
This is an example of how to assign specific labels to arbitrary positions
of a given axis using AxisItem.setTicks().

(Pure Qt)
"""

import sys
from pyqtgraph.Qt import QtGui
import pyqtgraph as pg
import numpy as np


if __name__ == "__main__":

    app = QtGui.QApplication([])

    w = pg.PlotWidget()

    # define a list of position,label tuples
    ticks = [list(zip((1, 2, 3, 7, 8), ("a", "b", "c", "d", "e")))]
    xax = w.getAxis("bottom")
    xax.setTicks(ticks)
    w.plot(np.arange(9))

    w.show()

    sys.exit(app.exec_())



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""Example on using a tpg.Y2ViewBox to provide a secondary Y axis"""

from PyQt5 import Qt
import pyqtgraph as pg
import numpy
from taurus_pyqtgraph import Y2ViewBox, CurvesPropertiesTool


if __name__ == "__main__":
    import sys

    app = Qt.QApplication([])

    w = pg.PlotWidget()

    # add Y2 viewbox (provides a ViewBox associated to bottom & right axes)
    y2 = Y2ViewBox()
    y2.attachToPlotItem(w.getPlotItem())

    # add a data item to Y1 (just as you would normally)
    c1 = pg.PlotDataItem(name="c1", pen="c")
    c1.setData(y=numpy.linspace(0, 20, 250))
    w.addItem(c1)

    # add a data item to Y2 (similar, but adding it to the secondary ViewBox!)
    c2 = pg.PlotDataItem(name="c2", pen="y")
    c2.setData(y=numpy.random.rand(250))
    y2.addItem(c2)  # <- note that it is y2, not w !

    # (optional) add CurvesPropertiesTool to switch curves between Y1 and Y2
    t = CurvesPropertiesTool()
    t.attachToPlotItem(w.getPlotItem(), y2=y2)

    w.show()
    sys.exit(app.exec_())
